import threading
import time
from collections import deque
from copy import deepcopy
from typing import (
    TYPE_CHECKING,
    Callable,
    Deque,
    Dict,
    List,
    Optional,
    Tuple,
    Type,
    Union,
)

from ingenialink.canopen.network import CanopenNetwork
from ingenialink.enums.register import RegCyclicType
from ingenialink.ethercat.network import EthercatNetwork
from ingenialink.ethercat.register import EthercatRegister
from ingenialink.ethercat.servo import EthercatServo
from ingenialink.exceptions import ILError, ILWrongWorkingCount
from ingenialink.pdo import RPDOMap, RPDOMapItem, TPDOMap, TPDOMapItem

from ingeniamotion.enums import COMMUNICATION_TYPE
from ingeniamotion.exceptions import IMException
from ingeniamotion.metaclass import DEFAULT_AXIS, DEFAULT_SERVO

if TYPE_CHECKING:
    from ingeniamotion.motion_controller import MotionController


class PDOPoller:
    """Poll register values using PDOs"""

    def __init__(
        self,
        mc: "MotionController",
        servo: str,
        refresh_time: float,
        watchdog_timeout: Optional[float],
        buffer_size: int,
    ) -> None:
        """Constructor.

        Args:
            mc: MotionController instance
            servo: drive alias.
            refresh_time: PDO values refresh time.
            watchdog_timeout: The PDO watchdog time. If not provided it will be set proportional
             to the refresh rate.
            buffer_size: Maximum number of register readings to store.

        """
        super().__init__()
        self.__mc = mc
        self.__servo = servo
        self.__refresh_time = refresh_time
        self.__watchdog_timeout = watchdog_timeout
        self.__buffer_size = buffer_size
        self.__buffer: List[Deque[Union[int, float]]] = []
        self.__timestamps: Deque[float] = deque(maxlen=self.__buffer_size)
        self.__start_time: Optional[float] = None
        self.__tpdo_map: TPDOMap = self.__mc.capture.pdo.create_empty_tpdo_map()
        self.__rpdo_map: RPDOMap = self.__mc.capture.pdo.create_empty_rpdo_map()
        self.__fill_rpdo_map()
        self.__exception_callbacks: List[Callable[[IMException], None]] = []

    def start(self) -> None:
        """Start the poller"""
        self.__mc.capture.pdo.set_pdo_maps_to_slave(
            self.__rpdo_map, self.__tpdo_map, servo=self.__servo
        )
        self.__mc.capture.pdo.subscribe_to_receive_process_data(self._new_data_available)
        for callback in self.__exception_callbacks:
            self.__mc.capture.pdo.subscribe_to_exceptions(callback)
        self.__start_time = time.time()
        self.__mc.capture.pdo.start_pdos(
            refresh_rate=self.__refresh_time, watchdog_timeout=self.__watchdog_timeout
        )

    def stop(self) -> None:
        """Stop the poller"""
        self.__mc.capture.pdo.stop_pdos()
        self.__mc.capture.pdo.unsubscribe_to_receive_process_data(self._new_data_available)
        for callback in self.__exception_callbacks:
            self.__mc.capture.pdo.unsubscribe_to_exceptions(callback)
        self.__mc.capture.pdo.remove_rpdo_map(self.__servo, self.__rpdo_map)
        self.__mc.capture.pdo.remove_tpdo_map(self.__servo, self.__tpdo_map)

    @property
    def data(self) -> Tuple[Deque[float], List[Deque[Union[int, float]]]]:
        """
        Get the poller data. After the data is retrieved, the data buffers are cleared.

        Returns:
            A tuple with a list of the readings timestamps and a list of lists with
            the readings values.

        """
        data = deepcopy(self.__buffer)
        time_stamps = deepcopy(self.__timestamps)
        self._clear_buffers()
        return time_stamps, data

    def add_channels(self, registers: List[Dict[str, Union[int, str]]]) -> None:
        """
        Configure the PDOs with the registers to be read.

        Args:
            registers : list of registers to add to the Poller.

        """
        self.__buffer = [deque(maxlen=self.__buffer_size) for _ in range(len(registers))]
        self.__fill_tpdo_map(registers)

    def subscribe_to_exceptions(self, callback: Callable[[IMException], None]) -> None:
        """Get notified when an exception occurs on the PDO thread.

        Args:
            callback: Function to be called when an exception occurs.

        """
        self.__exception_callbacks.append(callback)

    def _new_data_available(self) -> None:
        """Add readings to the buffers.

        Raises:
            ValueError: If the poller has not been started yet.

        """
        if self.__start_time is None:
            raise ValueError("The poller has not been started yet.")
        time_stamp = round(time.time() - self.__start_time, 6)
        self.__timestamps.append(time_stamp)
        for tpdo_index, tpdo_map_item in enumerate(self.__tpdo_map.items):
            self.__buffer[tpdo_index].append(tpdo_map_item.value)

    def _clear_buffers(self) -> None:
        """Clear the data buffers."""
        for buffer in self.__buffer:
            buffer.clear()
        self.__timestamps.clear()

    def __fill_rpdo_map(self) -> None:
        """Fill the RPDO Map with padding"""
        padding_rpdo_item = RPDOMapItem(size_bits=8)
        padding_rpdo_item.raw_data_bytes = int.to_bytes(0, 1, "little")
        self.__mc.capture.pdo.add_pdo_item_to_map(padding_rpdo_item, self.__rpdo_map)

    def __fill_tpdo_map(self, registers: List[Dict[str, Union[int, str]]]) -> None:
        """Fill the TPDO Map with the registers to be polled.

        Raises:
            ValueError: If there is a type mismatch when retrieving the register UID.
            ValueError: If there is a type mismatch when retrieving the register axis.

        """
        for register in registers:
            name = register.get("name", DEFAULT_SERVO)
            if not isinstance(name, str):
                raise ValueError(
                    f"Wrong type for the 'name' field. Expected 'str', got: {type(name)}"
                )
            axis = register.get("axis", DEFAULT_AXIS)
            if not isinstance(axis, int):
                raise ValueError(
                    f"Wrong type for the 'axis' field. Expected 'int', got: {type(axis)}"
                )
            tpdo_map_item = self.__mc.capture.pdo.create_pdo_item(
                name, axis=axis, servo=self.__servo
            )
            self.__mc.capture.pdo.add_pdo_item_to_map(tpdo_map_item, self.__tpdo_map)


class PDONetworkManager:
    """Manage all the PDO functionalities.

    Args:
        mc: The MotionController.

    """

    class ProcessDataThread(threading.Thread):
        """Manage the PDO exchange.

        Args:
            net: The EthercatNetwork instance where the PDOs will be active.
            refresh_rate: Determines how often (seconds) the PDO values will be updated.
            watchdog_timeout: The PDO watchdog time. If not provided it will be set proportional
             to the refresh rate.
            notify_send_process_data: Callback to notify when process data is about to be sent.
            notify_receive_process_data: Callback to notify when process data is received.
            notify_exceptions: Callback to notify when an exception is raised.

        Raises:
            ValueError: If the provided refresh rate is unfeasible.

        """

        DEFAULT_PDO_REFRESH_TIME = 0.01
        MINIMUM_PDO_REFRESH_TIME = 0.001
        MAXIMUM_PDO_REFRESH_TIME = 4
        DEFAULT_WATCHDOG_TIMEOUT = 0.1
        PDO_WATCHDOG_INCREMENT_FACTOR = 2
        # The time.sleep precision is 13 ms for Windows OS
        # https://stackoverflow.com/questions/1133857/how-accurate-is-pythons-time-sleep
        WINDOWS_TIME_SLEEP_PRECISION = 0.013

        def __init__(
            self,
            net: EthercatNetwork,
            refresh_rate: Optional[float],
            watchdog_timeout: Optional[float],
            notify_send_process_data: Optional[Callable[[], None]] = None,
            notify_receive_process_data: Optional[Callable[[], None]] = None,
            notify_exceptions: Optional[Callable[[IMException], None]] = None,
        ) -> None:
            super().__init__()
            self._net = net
            if refresh_rate is None:
                refresh_rate = self.DEFAULT_PDO_REFRESH_TIME
            elif refresh_rate < self.MINIMUM_PDO_REFRESH_TIME:
                raise ValueError(
                    f"The minimum PDO refresh rate is {self.MINIMUM_PDO_REFRESH_TIME} seconds."
                )
            elif refresh_rate > self.MAXIMUM_PDO_REFRESH_TIME:
                raise ValueError(
                    f"The maximum PDO refresh rate is {self.MAXIMUM_PDO_REFRESH_TIME} seconds."
                )
            self._refresh_rate = refresh_rate
            if watchdog_timeout is None:
                watchdog_timeout = max(
                    self.DEFAULT_WATCHDOG_TIMEOUT,
                    self._refresh_rate * self.PDO_WATCHDOG_INCREMENT_FACTOR,
                )
            self._watchdog_timeout = watchdog_timeout
            for servo in self._net.servos:
                servo.set_pdo_watchdog_time(watchdog_timeout)
            self._notify_send_process_data = notify_send_process_data
            self._notify_receive_process_data = notify_receive_process_data
            self._notify_exceptions = notify_exceptions
            self._pd_thread_stop_event = threading.Event()

        def run(self) -> None:
            """Start the PDO exchange"""
            try:
                self._net.start_pdos()
            except ILError as il_error:
                self._pd_thread_stop_event.set()
                if self._notify_exceptions is not None:
                    im_exception = IMException(
                        f"Could not start the PDOs due to the following exception: {il_error}"
                    )
                    self._notify_exceptions(im_exception)
            iteration_duration: float = -1
            while not self._pd_thread_stop_event.is_set():
                time_start = time.perf_counter()
                if self._notify_send_process_data is not None:
                    self._notify_send_process_data()
                try:
                    self._net.send_receive_processdata(self._refresh_rate)
                except ILWrongWorkingCount as il_error:
                    self._pd_thread_stop_event.set()
                    self._net.stop_pdos()
                    if iteration_duration == -1:
                        iteration_duration = time.perf_counter() - time_start
                    duration_error = ""
                    if iteration_duration > self._watchdog_timeout:
                        duration_error = (
                            f"Last iteration took {iteration_duration * 1000:0.1f} ms which is higher"
                            f" than the watchdog timeout ({self._watchdog_timeout * 1000:0.1f} ms). Please"
                            " optimize the callbacks and/or increase the refresh rate/watchdog timeout."
                        )
                    if self._notify_exceptions is not None:
                        im_exception = IMException(
                            "Stopping the PDO thread due to the following exception:"
                            f" {il_error} {duration_error}"
                        )
                        self._notify_exceptions(im_exception)
                else:
                    if self._notify_receive_process_data is not None:
                        self._notify_receive_process_data()
                    while (
                        remaining_loop_time := self._refresh_rate
                        - (time.perf_counter() - time_start)
                    ) > 0:
                        if remaining_loop_time > self.WINDOWS_TIME_SLEEP_PRECISION:
                            time.sleep(self.WINDOWS_TIME_SLEEP_PRECISION)
                        else:
                            self.high_precision_sleep(remaining_loop_time)
                    iteration_duration = time.perf_counter() - time_start

        def stop(self) -> None:
            """Stop the PDO exchange"""
            self._pd_thread_stop_event.set()
            self._net.stop_pdos()
            self.join()

        @staticmethod
        def high_precision_sleep(duration: float) -> None:
            """Replaces the time.sleep() method in order to obtain
            more precise sleeping times."""
            start_time = time.perf_counter()
            while duration - (time.perf_counter() - start_time) > 0:
                pass

    def __init__(self, motion_controller: "MotionController") -> None:
        self.mc = motion_controller
        self._pdo_thread: Optional[PDONetworkManager.ProcessDataThread] = None
        self._pdo_send_observers: List[Callable[[], None]] = []
        self._pdo_receive_observers: List[Callable[[], None]] = []
        self._pdo_exceptions_observers: List[Callable[[IMException], None]] = []

    def create_pdo_item(
        self,
        register_uid: str,
        axis: int = DEFAULT_AXIS,
        servo: str = DEFAULT_SERVO,
        value: Optional[Union[int, float]] = None,
    ) -> Union[RPDOMapItem, TPDOMapItem]:
        """
        Create a PDOMapItem by specifying a register UID.

        Args:
            register_uid: Register to be mapped.
            axis: servo axis. ``1`` by default.
            servo: servo alias to reference it. ``default`` by default.
            value: Initial value for an RPDO register.

        Returns:
            Mappable PDO item.

        Raises:
            ValueError: If there is a type mismatch retrieving the register object.
            AttributeError: If an initial value is not provided for an RPDO register.

        """
        pdo_map_item_dict: Dict[RegCyclicType, Type[Union[RPDOMapItem, TPDOMapItem]]] = {
            RegCyclicType.RX: RPDOMapItem,
            RegCyclicType.TX: TPDOMapItem,
        }
        drive = self.mc._get_drive(servo)
        register = drive.dictionary.registers(axis)[register_uid]
        if not isinstance(register, EthercatRegister):
            raise ValueError(f"Expected EthercatRegister. Got {type(register)}")
        pdo_map_item = pdo_map_item_dict[register.cyclic](register)
        if isinstance(pdo_map_item, RPDOMapItem):
            if value is None:
                raise AttributeError("A initial value is required for a RPDO.")
            pdo_map_item.value = value
        return pdo_map_item

    def create_pdo_maps(
        self,
        rpdo_map_items: Union[RPDOMapItem, List[RPDOMapItem]],
        tpdo_map_items: Union[TPDOMapItem, List[TPDOMapItem]],
    ) -> Tuple[RPDOMap, TPDOMap]:
        """
        Create the RPDO and TPDO maps from PDOMapItems.

        Args:
            rpdo_map_items: The RPDOMapItems to be added to a RPDOMap.
            tpdo_map_items: The TDOMapItems to be added to a TPDOMap.

        Returns:
            RPDO and TPDO maps.

        """
        rpdo_map = self.create_empty_rpdo_map()
        tpdo_map = self.create_empty_tpdo_map()
        if not isinstance(rpdo_map_items, list):
            rpdo_map_items = [rpdo_map_items]
        if not isinstance(tpdo_map_items, list):
            tpdo_map_items = [tpdo_map_items]
        for rpdo_map_item in rpdo_map_items:
            self.add_pdo_item_to_map(rpdo_map_item, rpdo_map)
        for tpdo_map_item in tpdo_map_items:
            self.add_pdo_item_to_map(tpdo_map_item, tpdo_map)
        return rpdo_map, tpdo_map

    @staticmethod
    def add_pdo_item_to_map(
        pdo_map_item: Union[RPDOMapItem, TPDOMapItem],
        pdo_map: Union[RPDOMap, TPDOMap],
    ) -> None:
        """
        Add a PDOMapItem to a PDOMap.

        Args:
            pdo_map_item: The PDOMapItem.
            pdo_map: The PDOMap to add the PDOMapItem.

        Raises:
            ValueError: If an RPDOItem is tried to be added to a TPDOMap.
            ValueError: If an TPDOItem is tried to be added to a RPDOMap.

        """
        if isinstance(pdo_map_item, RPDOMapItem) and not isinstance(pdo_map, RPDOMap):
            raise ValueError("Cannot add a RPDOItem to a TPDOMap")
        if isinstance(pdo_map_item, TPDOMapItem) and not isinstance(pdo_map, TPDOMap):
            raise ValueError("Cannot add a TPDOItem to a RPDOMap")
        pdo_map.add_item(pdo_map_item)

    @staticmethod
    def create_empty_rpdo_map() -> RPDOMap:
        """
        Create an empty RPDOMap.

        Returns:
            The empty RPDOMap.

        """
        return RPDOMap()

    @staticmethod
    def create_empty_tpdo_map() -> TPDOMap:
        """
        Create an empty TPDOMap.

        Returns:
            The empty TPDOMap.

        """
        return TPDOMap()

    def set_pdo_maps_to_slave(
        self,
        rpdo_maps: Union[RPDOMap, List[RPDOMap]],
        tpdo_maps: Union[TPDOMap, List[TPDOMap]],
        servo: str = DEFAULT_SERVO,
    ) -> None:
        """
        Map the PDOMaps to the slave.

        Args:
            rpdo_maps: The RPDOMaps to be mapped.
            tpdo_maps: he TPDOMaps to be mapped.
            servo: servo alias to reference it. ``default`` by default.

        Raises:
            ValueError: If there is a type mismatch retrieving the drive object.
            ValueError: If not all elements of the RPDO map list are instances of a RPDO map.
            ValueError: If not all elements of the TPDO map list are instances of a TPDO map.

        """
        drive = self.mc._get_drive(servo)
        if not isinstance(drive, EthercatServo):
            raise ValueError(f"Expected an EthercatServo. Got {type(drive)}")
        if not isinstance(rpdo_maps, list):
            rpdo_maps = [rpdo_maps]
        if not isinstance(tpdo_maps, list):
            tpdo_maps = [tpdo_maps]
        if not all(isinstance(rpdo_map, RPDOMap) for rpdo_map in rpdo_maps):
            raise ValueError("Not all elements of the RPDO map list are instances of a RPDO map")
        if not all(isinstance(tpdo_map, TPDOMap) for tpdo_map in tpdo_maps):
            raise ValueError("Not all elements of the TPDO map list are instances of a TPDO map")
        drive.set_pdo_map_to_slave(rpdo_maps, tpdo_maps)

    def clear_pdo_mapping(self, servo: str = DEFAULT_SERVO) -> None:
        """
        Clear the PDO mapping within the servo.

        Args:
            servo: servo alias to reference it. ``default`` by default.

        Raises:
            ValueError: If there is a type mismatch retrieving the drive object.

        """
        drive = self.mc._get_drive(servo)
        if not isinstance(drive, EthercatServo):
            raise ValueError(f"Expected an EthercatServo. Got {type(drive)}")
        drive.reset_rpdo_mapping()
        drive.reset_tpdo_mapping()

    def remove_rpdo_map(
        self,
        servo: str = DEFAULT_SERVO,
        rpdo_map: Optional[RPDOMap] = None,
        rpdo_map_index: Optional[int] = None,
    ) -> None:
        """Remove a RPDOMap from the RPDOMap list. The RPDOMap instance or
        the index of the map in the RPDOMap list should be provided.

        Args:
            servo: servo alias to reference it. ``default`` by default.
            rpdo_map: The RPDOMap instance to be removed.
            rpdo_map_index: The index of the RPDOMap list to be removed.

        Raises:
            ValueError: If the RPDOMap instance is not in the RPDOMap list.
            IndexError: If the index is out of range.

        """
        drive = self.mc._get_drive(servo)
        if not isinstance(drive, EthercatServo):
            raise ValueError(f"Expected an EthercatServo. Got {type(drive)}")
        drive.remove_rpdo_map(rpdo_map, rpdo_map_index)

    def remove_tpdo_map(
        self,
        servo: str = DEFAULT_SERVO,
        tpdo_map: Optional[TPDOMap] = None,
        tpdo_map_index: Optional[int] = None,
    ) -> None:
        """Remove a TPDOMap from the TPDOMap list. The TPDOMap instance or
        the index of the map in the TPDOMap list should be provided.

        Args:
            servo: servo alias to reference it. ``default`` by default.
            tpdo_map: The TPDOMap instance to be removed.
            tpdo_map_index: The index of the TPDOMap list to be removed.

        Raises:
            ValueError: If the TPDOMap instance is not in the TPDOMap list.
            IndexError: If the index is out of range.

        """
        drive = self.mc._get_drive(servo)
        if not isinstance(drive, EthercatServo):
            raise ValueError(f"Expected an EthercatServo. Got {type(drive)}")
        drive.remove_tpdo_map(tpdo_map, tpdo_map_index)

    def start_pdos(
        self,
        network_type: Optional[COMMUNICATION_TYPE] = None,
        refresh_rate: Optional[float] = None,
        watchdog_timeout: Optional[float] = None,
    ) -> None:
        """
        Start the PDO exchange process.

        Args:
            network_type: Network type (EtherCAT or CANopen) on which to start the PDO exchange.
            refresh_rate: Determines how often (seconds) the PDO values will be updated.
            watchdog_timeout: The PDO watchdog time. If not provided it will be set proportional
             to the refresh rate.

        Raises:
            ValueError: If the refresh rate is too high.
            ValueError: If the MotionController is connected to more than one Network.
            ValueError: If network_type argument is invalid.
            IMException: If the MotionController is connected to more than one Network.
            ValueError: If there is a type mismatch retrieving the network object.
            IMException: If the PDOs are already active.

        """
        if network_type is None:
            if len(self.mc.net) > 1:
                raise ValueError(
                    "There is more than one network created. The network_type argument must be"
                    " provided."
                )
            net = next(iter(self.mc.net.values()))
        elif not isinstance(network_type, COMMUNICATION_TYPE):
            raise ValueError(
                f"Wrong value for the network_type argument. Must be of type {COMMUNICATION_TYPE}"
            )
        elif network_type == COMMUNICATION_TYPE.Canopen:
            raise NotImplementedError
        else:
            ethercat_networks = [
                network for network in self.mc.net.values() if isinstance(network, EthercatNetwork)
            ]
            canopen_networks = [
                network for network in self.mc.net.values() if isinstance(network, CanopenNetwork)
            ]
            if len(ethercat_networks) > 1 or len(canopen_networks) > 1:
                raise IMException(
                    "When using PDOs only one instance per network type is allowed. "
                    f"Got {len(ethercat_networks)} instances of EthercatNetwork "
                    f"and {len(canopen_networks)} of CanopenNetwork."
                )
            net = (
                ethercat_networks[0]
                if network_type == COMMUNICATION_TYPE.Ethercat
                else canopen_networks[0]
            )
        if self._pdo_thread is not None:
            self.stop_pdos()
            raise IMException("PDOs are already active.")
        if not isinstance(net, EthercatNetwork):
            raise ValueError(f"Expected EthercatNetwork. Got {type(net)}")
        self._pdo_thread = self.ProcessDataThread(
            net,
            refresh_rate,
            watchdog_timeout,
            self._notify_send_process_data,
            self._notify_receive_process_data,
            self._notify_exceptions,
        )
        self._pdo_thread.start()

    def stop_pdos(self) -> None:
        """
        Stop the PDO exchange process.

        Raises:
            IMException: If the PDOs are not active yet.

        """
        if self._pdo_thread is None:
            raise IMException("The PDO exchange has not started yet.")
        self._pdo_thread.stop()
        self._pdo_thread = None

    def subscribe_to_send_process_data(self, callback: Callable[[], None]) -> None:
        """Subscribe be notified when the RPDO values will be sent.

        Args:
            callback: Callback function.

        """
        if callback in self._pdo_send_observers:
            return
        self._pdo_send_observers.append(callback)

    def subscribe_to_receive_process_data(self, callback: Callable[[], None]) -> None:
        """Subscribe be notified when the TPDO values are received.

        Args:
            callback: Callback function.

        """
        if callback in self._pdo_receive_observers:
            return
        self._pdo_receive_observers.append(callback)

    def subscribe_to_exceptions(self, callback: Callable[[IMException], None]) -> None:
        """Subscribe be notified when there is an exception in the PDO process data thread.

        If a callback is subscribed, the PDO exchange process is paused when an exception is raised.
        It can be resumed using the `resume_pdos` method.

        Args:
            callback: Callback function.

        """
        if callback in self._pdo_exceptions_observers:
            return
        self._pdo_exceptions_observers.append(callback)

    def unsubscribe_to_send_process_data(self, callback: Callable[[], None]) -> None:
        """Unsubscribe from the send process data notifications.

        Args:
            callback: Subscribed callback function.

        """
        if callback not in self._pdo_send_observers:
            return
        self._pdo_send_observers.remove(callback)

    def unsubscribe_to_receive_process_data(self, callback: Callable[[], None]) -> None:
        """Unsubscribe from the receive process data notifications.

        Args:
            callback: Subscribed callback function.

        """
        if callback not in self._pdo_receive_observers:
            return
        self._pdo_receive_observers.remove(callback)

    def create_poller(
        self,
        registers: List[Dict[str, Union[int, str]]],
        servo: str = DEFAULT_SERVO,
        sampling_time: float = 0.125,
        buffer_size: int = 100,
        watchdog_timeout: Optional[float] = None,
        start: bool = True,
    ) -> PDOPoller:
        """
        Create a register Poller using PDOs.

        Args:
            registers : list of registers to add to the Poller.
                Dicts should have the follow format:

                .. code-block:: python

                    [
                        { # Poller register one
                            "name": "CL_POS_FBK_VALUE",  # Register name.
                            "axis": 1  # Register axis.
                            # If it has no axis field, by default axis 1.
                        },
                        { # Poller register two
                            "name": "CL_VEL_FBK_VALUE",  # Register name.
                            "axis": 1  # Register axis.
                            # If it has no axis field, by default axis 1.
                        }
                    ]

            servo: servo alias to reference it. ``default`` by default.
            sampling_time: period of the sampling in seconds.
                By default ``0.125`` seconds.
            watchdog_timeout: The PDO watchdog time. If not provided it will be set proportional
             to the refresh rate.
            buffer_size: number maximum of sample for each data read.
                ``100`` by default.
            start: if ``True``, function starts poller, if ``False``
                poller should be started after. ``True`` by default.

        Returns:
            The poller instance.

        """
        poller = PDOPoller(self.mc, servo, sampling_time, watchdog_timeout, buffer_size)
        poller.add_channels(registers)
        if start:
            poller.start()
        return poller

    def unsubscribe_to_exceptions(self, callback: Callable[[IMException], None]) -> None:
        """Unsubscribe from the exceptions in the process data notifications.

        Args:
            callback: Subscribed callback function.

        """
        if callback not in self._pdo_exceptions_observers:
            return
        self._pdo_exceptions_observers.remove(callback)

    def _notify_send_process_data(self) -> None:
        """Notify subscribers that the RPDO values will be sent."""
        for callback in self._pdo_send_observers:
            callback()

    def _notify_receive_process_data(self) -> None:
        """Notify subscribers that the TPDO values were received."""
        for callback in self._pdo_receive_observers:
            callback()

    def _notify_exceptions(self, exc: IMException) -> None:
        """Notify subscribers that there were an exception.

        Args:
            exc: Exception that was raised in the PDO process data thread.
        """
        for callback in self._pdo_exceptions_observers:
            callback(exc)



from enum import IntEnum
from typing import TYPE_CHECKING, List, Optional, Tuple

if TYPE_CHECKING:
    from ingeniamotion.motion_controller import MotionController

from ingeniamotion.metaclass import MCMetaClass, DEFAULT_AXIS, DEFAULT_SERVO


class Errors(metaclass=MCMetaClass):
    """Errors."""

    class ErrorLocation(IntEnum):
        COCO = 0
        MOCO = 1
        SYSTEM = 2

    LAST_ERROR_COCO_REGISTER = "DRV_DIAG_ERROR_LAST_COM"
    LAST_ERROR_MOCO_REGISTER = "DRV_DIAG_ERROR_LAST"
    LAST_ERROR_SYSTEM_REGISTER = "DRV_DIAG_SYS_ERROR_LAST"
    LAST_ERROR_REGISTER = {
        ErrorLocation.COCO: LAST_ERROR_COCO_REGISTER,
        ErrorLocation.MOCO: LAST_ERROR_MOCO_REGISTER,
        ErrorLocation.SYSTEM: LAST_ERROR_SYSTEM_REGISTER,
    }
    ERROR_TOTAL_NUMBER_COCO_REGISTER = "DRV_DIAG_ERROR_TOTAL_COM"
    ERROR_TOTAL_NUMBER_MOCO_REGISTER = "DRV_DIAG_ERROR_TOTAL"
    ERROR_TOTAL_NUMBER_SYSTEM_REGISTER = "DRV_DIAG_SYS_ERROR_TOTAL_COM"
    ERROR_TOTAL_NUMBER_REGISTER = {
        ErrorLocation.COCO: ERROR_TOTAL_NUMBER_COCO_REGISTER,
        ErrorLocation.MOCO: ERROR_TOTAL_NUMBER_MOCO_REGISTER,
        ErrorLocation.SYSTEM: ERROR_TOTAL_NUMBER_SYSTEM_REGISTER,
    }
    ERROR_LIST_INDEX_REQUEST_COCO_REGISTER = "DRV_DIAG_ERROR_LIST_IDX_COM"
    ERROR_LIST_INDEX_REQUEST_MOCO_REGISTER = "DRV_DIAG_ERROR_LIST_IDX"
    ERROR_LIST_INDEX_REQUEST_SYSTEM_REGISTER = "DRV_DIAG_SYS_ERROR_LIST_IDX_COM"
    ERROR_LIST_INDEX_REQUEST_REGISTER = {
        ErrorLocation.COCO: ERROR_LIST_INDEX_REQUEST_COCO_REGISTER,
        ErrorLocation.MOCO: ERROR_LIST_INDEX_REQUEST_MOCO_REGISTER,
        ErrorLocation.SYSTEM: ERROR_LIST_INDEX_REQUEST_SYSTEM_REGISTER,
    }
    ERROR_LIST_REQUESTED_COCO_CODE = "DRV_DIAG_ERROR_LIST_CODE_COM"
    ERROR_LIST_REQUESTED_MOCO_CODE = "DRV_DIAG_ERROR_LIST_CODE"
    ERROR_LIST_REQUESTED_SYSTEM_CODE = "DRV_DIAG_SYS_ERROR_LIST_CODE_COM"
    ERROR_LIST_REQUESTED_CODE = {
        ErrorLocation.COCO: ERROR_LIST_REQUESTED_COCO_CODE,
        ErrorLocation.MOCO: ERROR_LIST_REQUESTED_MOCO_CODE,
        ErrorLocation.SYSTEM: ERROR_LIST_REQUESTED_SYSTEM_CODE,
    }

    MAXIMUM_ERROR_INDEX = 32

    STATUS_WORD_FAULT_BIT = 0x08
    STATUS_WORD_WARNING_BIT = 0x80

    ERROR_CODE_BITS = 0xFFFF
    ERROR_SUBNODE_BITS = 0xF00000
    ERROR_SUBNODE_SHIFT = 20
    ERROR_WARNING_BIT = 0x10000000
    ERROR_WARNING_SHIFT = 28

    def __init__(self, motion_controller: "MotionController") -> None:
        self.mc = motion_controller

    def __parse_error_to_tuple(
        self, error: int, location: ErrorLocation, subnode: Optional[int] = None
    ) -> Tuple[int, Optional[int], Optional[bool]]:
        error_code = error & self.ERROR_CODE_BITS
        if error_code == 0:
            return error_code, None, None
        if subnode is None:
            if location == self.ErrorLocation.MOCO:
                subnode = DEFAULT_AXIS
            else:
                subnode = (error & self.ERROR_SUBNODE_BITS) >> self.ERROR_SUBNODE_SHIFT
        is_warning = (error & self.ERROR_WARNING_BIT) >> self.ERROR_WARNING_SHIFT
        return error_code, subnode, bool(is_warning)

    def __get_error_location(self, servo: str = DEFAULT_SERVO) -> ErrorLocation:
        if self.mc.info.register_exists(self.LAST_ERROR_SYSTEM_REGISTER, axis=0, servo=servo):
            # Check System last error, if it does not exist check to CoCo
            return self.ErrorLocation.SYSTEM
        if self.mc.info.register_exists(self.LAST_ERROR_COCO_REGISTER, axis=0, servo=servo):
            # Check CoCo last error, if it does not exist use MoCo
            return self.ErrorLocation.COCO
        return self.ErrorLocation.MOCO

    def __get_error_subnode(
        self, location: ErrorLocation, subnode: Optional[int]
    ) -> Tuple[int, ErrorLocation]:
        if location == self.ErrorLocation.SYSTEM:
            if subnode is None:
                return 0, location
            elif subnode == 0:
                location = self.ErrorLocation.COCO
            elif subnode > 0:
                location = self.ErrorLocation.MOCO
            return subnode, location
        elif location == self.ErrorLocation.MOCO:
            return subnode or DEFAULT_AXIS, self.ErrorLocation.MOCO
        elif location == self.ErrorLocation.COCO:
            return 0, self.ErrorLocation.COCO

    def get_last_error(
        self, servo: str = DEFAULT_SERVO, axis: Optional[int] = None
    ) -> Tuple[int, Optional[int], Optional[bool]]:
        """Return last servo error.

        Args:
            servo : servo alias to reference it. ``default`` by default.
            axis : axis force read errors in target axis. ``None`` by default.

        Returns:
            Returns error data.

            code (int):
                Code error.
            axis (int):
                Error axis.
            is_warning (bool):
                ``True`` if warning, else ``False``.

        Raises:
            TypeError: If some read value has a wrong type.

        """
        error_version = self.__get_error_location(servo)
        subnode, error_location = self.__get_error_subnode(error_version, axis)
        error = self.mc.communication.get_register(
            self.LAST_ERROR_REGISTER[error_location], servo=servo, axis=subnode
        )
        if not isinstance(error, int):
            raise TypeError("Error has to be an integer")
        return self.__parse_error_to_tuple(error, error_version, axis)

    def get_last_buffer_error(
        self, servo: str = DEFAULT_SERVO, axis: Optional[int] = None
    ) -> Tuple[int, Optional[int], Optional[bool]]:
        """Get error code from error buffer last position.

        Args:
            servo : servo alias to reference it. ``default`` by default.
            axis : axis force read errors in target axis. ``None`` by default.

        Returns:
            Returns error data.

            code (int):
                Code error.
            axis (int):
                Error axis.
            is_warning (bool):
                ``True`` if warning, else ``False``.
        Raises:
            ValueError: Index must be less than 32
        """
        return self.get_buffer_error_by_index(0, servo=servo, axis=axis)

    def get_buffer_error_by_index(
        self, index: int, servo: str = DEFAULT_SERVO, axis: Optional[int] = None
    ) -> Tuple[int, Optional[int], Optional[bool]]:
        """Get error code from buffer error target index.

        Args:
            index : buffer error index. It must be less than ``32``.
            servo : servo alias to reference it. ``default`` by default.
            axis : axis force read errors in target axis. ``None`` by default.

        Returns:
            Returns error data.

            code (int):
                Code error.
            axis (int):
                Error axis.
            is_warning (bool):
                ``True`` if warning, else ``False``.
        Raises:
            ValueError: Index must be less than 32
            TypeError: If some read value has a wrong type.
        """
        if index >= self.MAXIMUM_ERROR_INDEX:
            raise ValueError("index must be less than 32")
        error_version = self.__get_error_location(servo)
        subnode, error_location = self.__get_error_subnode(error_version, axis)
        self.mc.communication.set_register(
            self.ERROR_LIST_INDEX_REQUEST_REGISTER[error_location], index, servo=servo, axis=subnode
        )
        error = self.mc.communication.get_register(
            self.ERROR_LIST_REQUESTED_CODE[error_location], servo=servo, axis=subnode
        )
        if not isinstance(error, int):
            raise TypeError("Error has to be an integer")
        return self.__parse_error_to_tuple(error, error_version, axis)

    def get_number_total_errors(
        self, servo: str = DEFAULT_SERVO, axis: Optional[int] = None
    ) -> int:
        """Return total number of drive errors.

        Args:
            servo : servo alias to reference it. ``default`` by default.
            axis : axis force read errors in target axis. ``None`` by default.

        Returns:
            Total number of errors.

        Raises:
            TypeError: If some read value has a wrong type.

        """
        error_version = self.__get_error_location(servo)
        subnode, error_location = self.__get_error_subnode(error_version, axis)
        total_number_errors = self.mc.communication.get_register(
            self.ERROR_TOTAL_NUMBER_REGISTER[error_location], servo=servo, axis=subnode
        )
        if not isinstance(total_number_errors, int):
            raise TypeError("Total number errors value has to be an integer")
        return total_number_errors

    def get_all_errors(
        self, servo: str = DEFAULT_SERVO, axis: Optional[int] = None
    ) -> List[Tuple[int, Optional[int], Optional[bool]]]:
        """Return List with all error codes.

        Args:
            servo : servo alias to reference it. ``default`` by default.
            axis : axis force read errors in target axis. ``None`` by default.

        Returns:
            List of all errors.
        """
        err_list = []
        err_num = self.get_number_total_errors(servo, axis)
        err_num = min(err_num, self.MAXIMUM_ERROR_INDEX)
        for i in range(err_num):
            error = self.get_buffer_error_by_index(i, servo=servo, axis=axis)
            err_list.append(error)
        return err_list

    def is_fault_active(self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS) -> bool:
        """Return if fault is active.

        Args:
            servo : servo alias to reference it. ``default`` by default.
            axis : servo axis. ``1`` by default.

        Returns:
            ``True`` if fault is active, else ``False``.
        """
        status_word = self.mc.configuration.get_status_word(servo=servo, axis=axis)
        return bool(status_word & self.STATUS_WORD_FAULT_BIT)

    def is_warning_active(self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS) -> bool:
        """Return if warning is active.

        Args:
            servo : servo alias to reference it. ``default`` by default.
            axis : servo axis. ``1`` by default.

        Returns:
            ``True`` if warning is active, else ``False``.
        """
        status_word = self.mc.configuration.get_status_word(servo=servo, axis=axis)
        return bool(status_word & self.STATUS_WORD_WARNING_BIT)

    def get_error_data(
        self, error_code: int, servo: str = DEFAULT_SERVO
    ) -> Tuple[str, str, str, str]:
        """Return error info from target error_code.

        Args:
            error_code : target error code.
            servo : servo alias to reference it. ``default`` by default.

        Returns:
           Returns error info.

            id (str):
                Error Id
            affected_module (str):
                Error affected module
            error_type (str):
                Error type
            error_message (str):
                Error message
        Raises:
            KeyError: The error codes does not exist in the error's dictionary.

        """
        drive = self.mc.servos[servo]
        dictionary_errors = drive.errors[error_code & self.ERROR_CODE_BITS]
        return tuple(dictionary_errors)  # type: ignore



from typing import TYPE_CHECKING, Dict, List, Optional, Union

import numpy as np
from numpy.typing import NDArray
from ingenialink.exceptions import ILIOError
from ingenialink.poller import Poller
from ingenialink.dictionary import SubnodeType

from ingeniamotion.disturbance import Disturbance
from ingeniamotion.enums import (
    MonitoringProcessStage,
    MonitoringSoCConfig,
    MonitoringSoCType,
    MonitoringVersion,
)
from ingeniamotion.exceptions import (
    IMMonitoringError,
    IMRegisterNotExist,
    IMStatusWordError,
)
from ingeniamotion.metaclass import DEFAULT_AXIS, DEFAULT_SERVO, MCMetaClass
from ingeniamotion.monitoring.base_monitoring import Monitoring
from ingeniamotion.monitoring.monitoring_v1 import MonitoringV1
from ingeniamotion.monitoring.monitoring_v3 import MonitoringV3
from ingeniamotion.pdo import PDONetworkManager

if TYPE_CHECKING:
    from ingeniamotion.motion_controller import MotionController


class Capture(metaclass=MCMetaClass):
    """Capture."""

    DISTURBANCE_STATUS_REGISTER = "DIST_STATUS"
    DISTURBANCE_MAXIMUM_SAMPLE_SIZE_REGISTER = "DIST_MAX_SIZE"
    MONITORING_STATUS_REGISTER = "MON_DIST_STATUS"
    MONITORING_CURRENT_NUMBER_BYTES_REGISTER = "MON_CFG_BYTES_VALUE"
    MONITORING_MAXIMUM_SAMPLE_SIZE_REGISTER = "MON_MAX_SIZE"
    MONITORING_FREQUENCY_DIVIDER_REGISTER = "MON_DIST_FREQ_DIV"

    MINIMUM_BUFFER_SIZE = 8192

    MONITORING_VERSION_REGISTER = "MON_DIST_VERSION"

    MONITORING_STATUS_ENABLED_BIT = 0x1
    DISTURBANCE_STATUS_ENABLED_BIT = 0x1

    MONITORING_STATUS_PROCESS_STAGE_BITS = {
        MonitoringVersion.MONITORING_V1: 0x6,
        MonitoringVersion.MONITORING_V2: 0x6,
        MonitoringVersion.MONITORING_V3: 0xE,
    }
    MONITORING_AVAILABLE_FRAME_BIT = {
        MonitoringVersion.MONITORING_V1: 0x800,
        MonitoringVersion.MONITORING_V2: 0x800,
        MonitoringVersion.MONITORING_V3: 0x10,
    }

    def __init__(self, motion_controller: "MotionController") -> None:
        self.mc = motion_controller
        self.pdo = PDONetworkManager(self.mc)

    def create_poller(
        self,
        registers: List[Dict[str, Union[int, str]]],
        servo: str = DEFAULT_SERVO,
        sampling_time: float = 0.125,
        buffer_size: int = 100,
        start: bool = True,
    ) -> Poller:
        """Returns a Poller instance with target registers.

        Args:
            registers : list of registers to add to the Poller.
                Dicts should have the follow format:

                .. code-block:: python

                    [
                        { # Poller register one
                            "name": "CL_POS_FBK_VALUE",  # Register name.
                            "axis": 1  # Register axis.
                            # If it has no axis field, by default axis 1.
                        },
                        { # Poller register two
                            "name": "CL_VEL_FBK_VALUE",  # Register name.
                            "axis": 1  # Register axis.
                            # If it has no axis field, by default axis 1.
                        }
                    ]

            servo: servo alias to reference it. ``default`` by default.
            sampling_time: period of the sampling in seconds.
                By default ``0.125`` seconds.
            buffer_size: number maximum of sample for each data read.
                ``100`` by default.
            start: if ``True``, function starts poller, if ``False``
                poller should be started after. ``True`` by default.

        Returns:
            Poller object with chosen registers.

            Poller.start()
                Poller starts reading the registers.

            Poller.stop()
                Poller stop reading the registers.

            Poller.data
                tuple with 3 items: a list of timestamp, list of lists of values
                (one list of values for each register), and a boolean that
                indicates if data was lost.

                When the poller starts, the lists are filled with the timestamp
                and the value of the registers reading. The maximum length of
                the list will be buffer_size value, when this size is reached,
                the older value will be removed and the newest will be added.

                When the property data is read list are reset to a empty list.

        Raises:
            IMRegisterNotExist: If register does not exist in dictionary.
            TypeError: If some parameter has a wrong type.

        """
        poller = Poller(self.mc.servos[servo], len(registers))
        poller.configure(sampling_time, buffer_size)
        for index, register in enumerate(registers):
            axis = register.get("axis", DEFAULT_AXIS)
            name = register.get("name")
            if not isinstance(axis, int):
                raise TypeError("Wrong axis type, it should be an int")
            if not isinstance(name, str):
                raise TypeError("Name type is a string")
            register_obj = self.mc.info.register_info(name, axis, servo=servo)
            poller.ch_configure(index, register_obj)
        if start:
            poller.start()
        return poller

    def create_empty_monitoring(self, servo: str = DEFAULT_SERVO) -> Monitoring:
        """Returns a Monitoring instance not configured.

        Args:
            servo : servo alias to reference it. ``default`` by default.

        Returns:
            Not configured instance of monitoring.

        Raises:
            NotImplementedError: If an wrong monitoring version is requested.

        """
        version = self._check_version(servo)
        if version == MonitoringVersion.MONITORING_V3:
            return MonitoringV3(self.mc, servo)
        elif version == MonitoringVersion.MONITORING_V1:
            return MonitoringV1(self.mc, servo)
        else:
            raise NotImplementedError(f"This version {version} is not implemented yet")

    def create_monitoring(
        self,
        registers: List[Dict[str, Union[int, str]]],
        prescaler: int,
        sample_time: float,
        trigger_delay: float = 0,
        trigger_mode: MonitoringSoCType = MonitoringSoCType.TRIGGER_EVENT_AUTO,
        trigger_config: Optional[MonitoringSoCConfig] = None,
        trigger_signal: Optional[Dict[str, Union[int, str]]] = None,
        trigger_value: Union[float, int, None] = None,
        servo: str = DEFAULT_SERVO,
        start: bool = False,
    ) -> Monitoring:
        """Returns a Monitoring instance configured with target registers.

        Args:
            registers: list of registers to add to Monitoring.
                Dicts should have the follow format:

                .. code-block:: python

                    [
                        { # Monitoring register one
                            "name": "CL_POS_FBK_VALUE",  # Register name.
                            "axis": 1  # Register axis.
                            # If it has no axis field, by default axis 1.
                        },
                        { # Monitoring register two
                            "name": "CL_VEL_FBK_VALUE",  # Register name.
                            "axis": 1  # Register axis.
                            # If it has no axis field, by default axis 1.
                        }
                    ]

            prescaler : determines monitoring frequency. Frequency will be
                ``Position & velocity loop rate frequency / prescaler``, see
                :func:`ingeniamotion.configuration.Configuration.get_position_and_velocity_loop_rate`
                to know about this frequency. It must be ``1`` or higher.
            sample_time : sample time in seconds.
            trigger_delay : trigger delay in seconds. Value should be between
                ``-sample_time/2`` and ``sample_time/2`` . ``0`` by default.
            trigger_mode : monitoring start of condition type.
                ``TRIGGER_EVENT_NONE`` by default.
            trigger_config : monitoring edge condition.
                ``None`` by default.
            trigger_signal : dict with name and axis of trigger signal
                for rising or falling edge trigger.
            trigger_value : value for rising or falling edge trigger.
            servo : servo alias to reference it. ``default`` by default.
            start : if ``True``, function starts monitoring, if ``False``
                monitoring should be started after. ``False`` by default.

        Returns:
            Instance of monitoring configured.

        Raises:
            ValueError: If prescaler is less than ``1``.
            ValueError: If trigger_delay is not between ``-total_time/2`` and
             ``total_time/2``.
            IMMonitoringError: If register maps fails in the servo.
            IMMonitoringError: If buffer size is not enough for all the registers
             and samples.
            IMMonitoringError: If trigger_mode is rising or falling edge trigger
             and trigger signal is not mapped.
            TypeError: If trigger_mode is rising or falling edge trigger and
             trigger_signal or trigger_value are None.

        """
        self.clean_monitoring(servo=servo)
        monitoring = self.create_empty_monitoring(servo)
        monitoring.set_frequency(prescaler)
        monitoring.map_registers(registers)
        monitoring.set_trigger(
            trigger_mode,
            edge_condition=trigger_config,
            trigger_signal=trigger_signal,
            trigger_value=trigger_value,
        )
        monitoring.configure_sample_time(sample_time, trigger_delay)
        if start:
            self.enable_monitoring(servo=servo)
        return monitoring

    def create_disturbance(
        self,
        register: str,
        data: Union[List[Union[float, int]], NDArray[np.int_], NDArray[np.float_]],
        freq_divider: int,
        servo: str = DEFAULT_SERVO,
        axis: int = DEFAULT_AXIS,
        start: bool = False,
    ) -> Disturbance:
        """Returns a Disturbance instance configured with target registers.

        Args:
            register : target register UID.
            data : data to write in disturbance.
            freq_divider : determines disturbance frequency divider. Frequency will
                be ``Position & velocity loop rate frequency / freq_divider``, see
                :func:`ingeniamotion.configuration.Configuration.get_position_and_velocity_loop_rate`
                to know about this frequency. It must be ``1`` or higher.
            servo : servo alias to reference it. ``default`` by default.
            axis : servo axis. ``1`` by default.
            start : if ``True``, function starts disturbance,
                if ``False`` disturbance should be started after.
                ``False`` by default.

        Returns:
            Instance of disturbance configured.

        Raises:
            ValueError: If freq_divider is less than ``1``.
            IMDisturbanceError: If buffer size is not enough for all the
                registers and samples.

        """
        self.clean_disturbance(servo=servo)
        disturbance = Disturbance(self.mc, servo)
        disturbance.set_frequency_divider(freq_divider)
        disturbance.map_registers({"name": register, "axis": axis})
        disturbance.write_disturbance_data(data)
        if start:
            self.enable_disturbance(servo=servo)
        return disturbance

    def _check_version(self, servo: str) -> MonitoringVersion:
        """Checks the version of the monitoring based on a given servo.

        Args:
            servo : servo alias to reference it. ``default`` by default.

        Raises:
            NotImplementedError: If the drive does not support monitoring
            and disturbance.

        """
        drive = self.mc._get_drive(servo)
        try:
            self.mc.communication.get_register(
                self.MONITORING_VERSION_REGISTER, servo=servo, axis=0
            )
            return MonitoringVersion.MONITORING_V3
        except (IMRegisterNotExist, ILIOError):
            # The Monitoring V3 is NOT available
            pass
        try:
            self.mc.communication.get_register(
                self.MONITORING_CURRENT_NUMBER_BYTES_REGISTER, servo=servo, axis=0
            )
            return MonitoringVersion.MONITORING_V2
        except (IMRegisterNotExist, ILIOError):
            # The Monitoring V2 is NOT available
            pass
        try:
            self.mc.communication.get_register(self.MONITORING_STATUS_REGISTER, servo=servo, axis=0)
            return MonitoringVersion.MONITORING_V1
        except (IMRegisterNotExist, ILIOError):
            # Monitoring/disturbance are not available
            raise NotImplementedError(
                "The monitoring and disturbance features are not available for this drive"
            )

    def enable_monitoring_disturbance(self, servo: str = DEFAULT_SERVO) -> None:
        """Enable monitoring and disturbance.

        Args:
            servo : servo alias to reference it. ``default`` by default.

        Raises:
            IMMonitoringError: If monitoring can't be enabled.

        """
        self.enable_monitoring(servo=servo)
        self.enable_disturbance(servo=servo)

    def enable_monitoring(self, servo: str = DEFAULT_SERVO) -> None:
        """Enable monitoring.

        Args:
            servo : servo alias to reference it. ``default`` by default.

        Raises:
            IMMonitoringError: If monitoring can't be enabled.

        """
        drive = self.mc.servos[servo]
        drive.monitoring_enable()
        # Check monitoring status
        if not self.is_monitoring_enabled(servo=servo):
            raise IMMonitoringError("Error enabling monitoring.")

    def enable_disturbance(
        self, servo: str = DEFAULT_SERVO, version: Optional[MonitoringVersion] = None
    ) -> None:
        """Enable disturbance.

        Args:
            servo : servo alias to reference it. ``default`` by default.
            version : Monitoring/Disturbance version,
                if ``None`` reads from drive. ``None`` by default.

        Raises:
            IMMonitoringError: If disturbance can't be enabled.

        """
        if version is None:
            version = self._check_version(servo)
        if version < MonitoringVersion.MONITORING_V3:
            return self.enable_monitoring(servo=servo)
        drive = self.mc.servos[servo]
        drive.disturbance_enable()
        # Check disturbance status
        if not self.is_disturbance_enabled(servo=servo):
            raise IMMonitoringError("Error enabling disturbance.")

    def disable_monitoring_disturbance(self, servo: str = DEFAULT_SERVO) -> None:
        """Disable monitoring and disturbance.

        Args:
            servo : servo alias to reference it. ``default`` by default.

        """
        self.disable_monitoring(servo=servo)
        self.disable_disturbance(servo=servo)

    def disable_monitoring(
        self, servo: str = DEFAULT_SERVO, version: Optional[MonitoringVersion] = None
    ) -> None:
        """Disable monitoring.

        Args:
            servo : servo alias to reference it. ``default`` by default.
            version : Monitoring/Disturbance version,
                if ``None`` reads from drive. ``None`` by default.

        """
        if version is None:
            version = self._check_version(servo)
        if not self.is_monitoring_enabled(servo=servo):
            return
        drive = self.mc.servos[servo]
        drive.monitoring_disable()
        if version >= MonitoringVersion.MONITORING_V3:
            drive.monitoring_remove_data()

    def disable_disturbance(
        self, servo: str = DEFAULT_SERVO, version: Optional[MonitoringVersion] = None
    ) -> None:
        """Disable disturbance.

        Args:
            servo : servo alias to reference it. ``default`` by default.
            version : Monitoring/Disturbance version,
                if ``None`` reads from drive. ``None`` by default.

        """
        if version is None:
            version = self._check_version(servo)
        if not self.is_disturbance_enabled(servo, version):
            return
        if version < MonitoringVersion.MONITORING_V3:
            return self.disable_monitoring(servo=servo, version=version)
        drive = self.mc.servos[servo]
        drive.disturbance_disable()
        drive.disturbance_remove_data()

    def get_monitoring_disturbance_status(self, servo: str = DEFAULT_SERVO) -> int:
        """Get Monitoring Status.

        Args:
            servo : servo alias to reference it. ``default`` by default.

        Returns:
            Monitoring/Disturbance Status.

        Raises:
            IMRegisterNotExist: If the register doesn't exist.
            TypeError: If some read value has a wrong type.

        """
        monitoring_disturbance_status = self.mc.communication.get_register(
            self.MONITORING_STATUS_REGISTER, servo=servo, axis=0
        )
        if not isinstance(monitoring_disturbance_status, int):
            raise TypeError("Monitoring and disturbance status value has to be an integer")
        return monitoring_disturbance_status

    def get_monitoring_status(self, servo: str = DEFAULT_SERVO) -> int:
        """Get Monitoring Status.

        Args:
            servo : servo alias to reference it. ``default`` by default.

        Returns:
            Monitoring Status.

        Raises:
            IMRegisterNotExist: If the register doesn't exist.
            TypeError: If some read value has a wrong type.

        """
        monitoring_status = self.mc.communication.get_register(
            self.MONITORING_STATUS_REGISTER, servo=servo, axis=0
        )
        if not isinstance(monitoring_status, int):
            raise TypeError("Monitoring status value has to be an integer")
        return monitoring_status

    def get_disturbance_status(
        self, servo: str = DEFAULT_SERVO, version: Optional[MonitoringVersion] = None
    ) -> int:
        """Get Disturbance Status.

        Args:
            servo : servo alias to reference it. ``default`` by default.
            version : Monitoring/Disturbance version,
                if ``None`` reads from drive. ``None`` by default.

        Returns:
            Disturbance Status.

        Raises:
            IMRegisterNotExist: If the register doesn't exist.
            TypeError: If some read value has a wrong type.

        """
        if version is None:
            version = self._check_version(servo)
        if version < MonitoringVersion.MONITORING_V3:
            disturbance_status = self.mc.communication.get_register(
                self.MONITORING_STATUS_REGISTER, servo=servo, axis=0
            )
        else:
            disturbance_status = self.mc.communication.get_register(
                self.DISTURBANCE_STATUS_REGISTER, servo=servo, axis=0
            )
        if not isinstance(disturbance_status, int):
            raise TypeError("Disturbance status value has to be an integer")
        return disturbance_status

    def is_monitoring_enabled(self, servo: str = DEFAULT_SERVO) -> bool:
        """Check if monitoring is enabled.

        Args:
            servo : servo alias to reference it. ``default`` by default.

        Returns:
            True if monitoring is enabled, else False.

        Raises:
            IMRegisterNotExist: If the register doesn't exist.

        """
        monitor_status = self.get_monitoring_status(servo)
        return (monitor_status & self.MONITORING_STATUS_ENABLED_BIT) == 1

    def is_disturbance_enabled(
        self, servo: str = DEFAULT_SERVO, version: Optional[MonitoringVersion] = None
    ) -> bool:
        """Check if disturbance is enabled.

        Args:
            servo : servo alias to reference it. ``default`` by default.
            version : Monitoring/Disturbance version,
                if ``None`` reads from drive. ``None`` by default.

        Returns:
            True if disturbance is enabled, else False.

        Raises:
            IMRegisterNotExist: If the register doesn't exist.

        """
        monitor_status = self.get_disturbance_status(servo, version=version)
        return (monitor_status & self.DISTURBANCE_STATUS_ENABLED_BIT) == 1

    def get_monitoring_process_stage(
        self, servo: str = DEFAULT_SERVO, version: Optional[MonitoringVersion] = None
    ) -> MonitoringProcessStage:
        """
        Return monitoring process stage.

        Args:
            servo : servo alias to reference it. ``default`` by default.
            version : Monitoring/Disturbance version,
                if ``None`` reads from drive. ``None`` by default.

        Returns:
            Current monitoring process stage.

        Raises:
            IMRegisterNotExist: If the register doesn't exist.

        """
        if version is None:
            version = self._check_version(servo=servo)
        monitor_status = self.mc.capture.get_monitoring_status(servo=servo)
        mask = self.MONITORING_STATUS_PROCESS_STAGE_BITS[version]
        masked_value = monitor_status & mask
        return MonitoringProcessStage(masked_value)

    def is_frame_available(
        self, servo: str = DEFAULT_SERVO, version: Optional[MonitoringVersion] = None
    ) -> bool:
        """
        Check if monitoring has an available frame.

        Args:
            servo : servo alias to reference it. ``default`` by default.
            version : Monitoring/Disturbance version,
                if ``None`` reads from drive. ``None`` by default.

        Returns:
            True if monitoring has an available frame, else False.

        Raises:
            IMRegisterNotExist: If the register doesn't exist.

        """
        if version is None:
            version = self._check_version(servo=servo)
        monitor_status = self.mc.capture.get_monitoring_status(servo=servo)
        mask = self.MONITORING_AVAILABLE_FRAME_BIT[version]
        return (monitor_status & mask) != 0

    def clean_monitoring(
        self, servo: str = DEFAULT_SERVO, version: Optional[MonitoringVersion] = None
    ) -> None:
        """Disable monitoring/disturbance and remove monitoring mapped registers.

        Args:
            servo : servo alias to reference it. ``default`` by default.
            version : Monitoring/Disturbance version,
                if None reads from drive. ``None`` by default.

        """
        self.disable_monitoring(servo=servo, version=version)
        drive = self.mc.servos[servo]
        drive.monitoring_remove_all_mapped_registers()

    def clean_disturbance(
        self, servo: str = DEFAULT_SERVO, version: Optional[MonitoringVersion] = None
    ) -> None:
        """Disable monitoring/disturbance and remove disturbance mapped registers.

        Args:
            servo : servo alias to reference it. ``default`` by default.
            version : Monitoring/Disturbance version,
                if None reads from drive. ``None`` by default.

        """
        self.disable_disturbance(servo=servo, version=version)
        drive = self.mc.servos[servo]
        drive.disturbance_remove_all_mapped_registers()

    def clean_monitoring_disturbance(self, servo: str = DEFAULT_SERVO) -> None:
        """Disable monitoring/disturbance, remove disturbance and monitoring
        mapped registers.

        Args:
            servo : servo alias to reference it. ``default`` by default.

        """
        self.clean_monitoring(servo=servo)
        self.clean_disturbance(servo=servo)

    def mcb_synchronization(self, servo: str = DEFAULT_SERVO) -> None:
        """Synchronize MCB, necessary to monitoring and disturbance.
        Motor must be disabled.

        Args:
            servo : servo alias to reference it. ``default`` by default.

        Raises:
            IMStatusWordError: If motor is enabled.

        """
        for subnode in [
            subnode
            for subnode, subnode_type in self.mc.info.get_subnodes(servo).items()
            if subnode_type == SubnodeType.MOTION
        ]:
            if self.mc.configuration.is_motor_enabled(servo=servo, axis=subnode):
                raise IMStatusWordError("Motor is enabled")
        self.enable_monitoring(servo=servo)
        self.disable_monitoring(servo=servo)

    def disturbance_max_sample_size(self, servo: str = DEFAULT_SERVO) -> int:
        """Return disturbance max size, in bytes.

        Args:
            servo : servo alias to reference it. ``default`` by default.

        Returns:
            Max buffer size in bytes.

        Raises:
            TypeError: If some read value has a wrong type.

        """
        try:
            max_sample_size = self.mc.communication.get_register(
                self.DISTURBANCE_MAXIMUM_SAMPLE_SIZE_REGISTER, servo=servo, axis=0
            )
            if not isinstance(max_sample_size, int):
                raise TypeError("Maximum sample size has to be an integer")
            return max_sample_size
        except IMRegisterNotExist:
            return self.MINIMUM_BUFFER_SIZE

    def monitoring_max_sample_size(self, servo: str = DEFAULT_SERVO) -> int:
        """Return monitoring max size, in bytes.

        Args:
            servo : servo alias to reference it. ``default`` by default.

        Returns:
            Max buffer size in bytes.

        Raises:
            TypeError: If some read value has a wrong type.

        """
        try:
            max_sample_size = self.mc.communication.get_register(
                self.MONITORING_MAXIMUM_SAMPLE_SIZE_REGISTER, servo=servo, axis=0
            )
            if not isinstance(max_sample_size, int):
                raise TypeError("Maximum sample size has to be an integer")
            return max_sample_size
        except IMRegisterNotExist:
            return self.MINIMUM_BUFFER_SIZE

    def get_frequency(self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS) -> float:
        """Returns the monitoring frequency.

        Args:
            servo: servo alias to reference it. ``default`` by default.
            axis: servo axis. ``1`` by default.

        Returns:
            Sampling rate in Hz.

        Raises:
            TypeError: If some read value has a wrong type.

        """

        position_velocity_loop_rate = self.mc.configuration.get_position_and_velocity_loop_rate(
            servo=servo, axis=axis
        )
        prescaler = self.mc.communication.get_register(
            self.MONITORING_FREQUENCY_DIVIDER_REGISTER, servo=servo, axis=0
        )
        if not isinstance(prescaler, int):
            raise TypeError("Monitoring loop frequency divider has to be an integer")
        sampling_freq = round(position_velocity_loop_rate / prescaler, 2)
        return sampling_freq



import pytest

from ingeniamotion.disturbance import Disturbance
from ingeniamotion.exceptions import IMDisturbanceError


@pytest.fixture
def disturbance_map_registers(disturbance, skip_if_monitoring_not_available):
    disturbance.map_registers([{"axis": 1, "name": "CL_TOR_SET_POINT_VALUE"}])


@pytest.fixture
def disturbance(motion_controller, skip_if_monitoring_not_available):
    mc, alias = motion_controller
    return Disturbance(mc, alias)


@pytest.mark.virtual
@pytest.mark.smoke
def test_disturbance_max_sample_size(motion_controller, disturbance):
    mc, alias = motion_controller
    max_sample_size = disturbance.max_sample_number
    value = mc.communication.get_register(
        disturbance.DISTURBANCE_MAXIMUM_SAMPLE_SIZE_REGISTER, servo=alias, axis=0
    )
    assert max_sample_size == value


@pytest.mark.smoke
@pytest.mark.parametrize("prescaler", list(range(2, 11, 2)))
def test_set_frequency_divider(motion_controller, disturbance, prescaler):
    mc, alias = motion_controller
    disturbance.set_frequency_divider(prescaler)
    value = mc.communication.get_register(
        disturbance.DISTURBANCE_FREQUENCY_DIVIDER_REGISTER, servo=alias, axis=0
    )
    assert value == prescaler


@pytest.mark.smoke
def test_set_frequency_divider_exception(disturbance):
    prescaler = -1
    with pytest.raises(ValueError):
        disturbance.set_frequency_divider(prescaler)


@pytest.mark.smoke
@pytest.mark.parametrize(
    "axis, name, expected_value",
    [
        (1, "CL_POS_SET_POINT_VALUE", 270533892),
        (1, "CL_VEL_SET_POINT_VALUE", 270600196),
        (1, "CL_TOR_SET_POINT_VALUE", 270665732),
    ],
)
def test_disturbance_map_registers(motion_controller, disturbance, axis, name, expected_value):
    mc, alias = motion_controller
    registers = [{"axis": axis, "name": name}]
    disturbance.map_registers(registers)
    value = mc.communication.get_register("DIST_CFG_REG0_MAP", servo=alias, axis=0)
    assert value == expected_value
    value = mc.communication.get_register("DIST_CFG_MAP_REGS", servo=alias, axis=0)
    assert value == 1


@pytest.mark.smoke
@pytest.mark.parametrize("number_registers", list(range(1, 17)))
def test_disturbance_number_map_registers(motion_controller, disturbance, number_registers):
    mc, alias = motion_controller
    reg_dict = {"axis": 1, "name": "CL_POS_SET_POINT_VALUE"}
    registers = [reg_dict for _ in range(number_registers)]
    disturbance.map_registers(registers)
    value = mc.communication.get_register("DIST_CFG_MAP_REGS", servo=alias, axis=0)
    assert value == number_registers


@pytest.mark.smoke
def test_disturbance_map_registers_sample_number(disturbance):
    registers = [{"axis": 1, "name": "CL_POS_SET_POINT_VALUE"}]
    value = disturbance.map_registers(registers)
    assert value == disturbance.max_sample_number / 4


@pytest.mark.virtual
@pytest.mark.smoke
def test_disturbance_map_registers_exception(disturbance):
    registers = [{"axis": 0, "name": "DRV_AXIS_NUMBER"}]
    with pytest.raises(IMDisturbanceError):
        disturbance.map_registers(registers)


@pytest.mark.virtual
@pytest.mark.smoke
def test_disturbance_map_registers_empty(disturbance):
    registers = []
    with pytest.raises(IMDisturbanceError):
        disturbance.map_registers(registers)


@pytest.mark.smoke
@pytest.mark.usefixtures("disturbance_map_registers")
def test_write_disturbance_data_buffer_exception(disturbance):
    with pytest.raises(IMDisturbanceError):
        disturbance.write_disturbance_data([0] * disturbance.max_sample_number)


@pytest.mark.virtual
@pytest.mark.smoke
def test_write_disturbance_data_not_configured(disturbance):
    with pytest.raises(IMDisturbanceError):
        disturbance.write_disturbance_data([0] * 100)


@pytest.mark.smoke
def test_write_disturbance_data_enabled(
    motion_controller, disturbance, disable_monitoring_disturbance
):
    mc, alias = motion_controller
    mc.capture.enable_monitoring_disturbance(alias)
    with pytest.raises(IMDisturbanceError):
        disturbance.write_disturbance_data([0] * 100)


@pytest.mark.virtual
@pytest.mark.smoke
def test_disturbance_map_registers_invalid_subnode(mocker, motion_controller, disturbance):
    registers = [{"axis": "1", "name": "DRV_AXIS_NUMBER"}]
    mc, alias = motion_controller
    mocker.patch.object(mc.capture, "is_disturbance_enabled", return_value=False)
    with pytest.raises(TypeError):
        disturbance.map_registers(registers)


@pytest.mark.virtual
@pytest.mark.smoke
def test_disturbance_map_registers_invalid_register(mocker, motion_controller, disturbance):
    registers = [{"axis": 1, "name": 1}]
    mc, alias = motion_controller
    mocker.patch.object(mc.capture, "is_disturbance_enabled", return_value=False)
    with pytest.raises(TypeError):
        disturbance.map_registers(registers)


@pytest.mark.virtual
@pytest.mark.smoke
def test_write_disturbance_data_wrong_data_type(mocker, motion_controller, disturbance):
    mc, alias = motion_controller
    mocker.patch.object(mc.capture, "is_disturbance_enabled", return_value=False)
    registers = [{"axis": 1, "name": "CL_POS_SET_POINT_VALUE"}]
    disturbance.map_registers(registers)
    mocker.patch.object(disturbance, "sampling_freq", return_value=1000)
    with pytest.raises(TypeError):
        disturbance.write_disturbance_data(["wrong", "input", "value"])



import time

import numpy as np
import pytest
from ingenialink import exceptions

from ingeniamotion.enums import OperationMode
from ingeniamotion.exceptions import IMTimeoutError
from ingeniamotion.motion import Motion
from tests.conftest import mean_actual_velocity_position

POS_PID_KP_VALUE = 0.1
POSITION_PERCENTAGE_ERROR_ALLOWED = 5

PROFILER_LATCHING_MODE_REGISTER = "PROF_LATCH_MODE"
OPERATION_MODE_REGISTER = "DRV_OP_CMD"
POSITION_SET_POINT_REGISTER = "CL_POS_SET_POINT_VALUE"
ACTUAL_POSITION_REGISTER = "CL_POS_FBK_VALUE"
VELOCITY_SET_POINT_REGISTER = "CL_VEL_SET_POINT_VALUE"
ACTUAL_VELOCITY_REGISTER = "CL_VEL_FBK_VALUE"
CURRENT_QUADRATURE_SET_POINT_REGISTER = "CL_CUR_Q_SET_POINT"
ACTUAL_QUADRATURE_CURRENT_REGISTER = "CL_CUR_Q_VALUE"
CURRENT_DIRECT_SET_POINT_REGISTER = "CL_CUR_D_SET_POINT"
ACTUAL_DIRECT_CURRENT_REGISTER = "CL_CUR_D_VALUE"
VOLTAGE_QUADRATURE_SET_POINT_REGISTER = "CL_VOL_Q_SET_POINT"
VOLTAGE_DIRECT_SET_POINT_REGISTER = "CL_VOL_D_SET_POINT"


def test_target_latch(motion_controller):
    mc, alias = motion_controller
    mc.communication.set_register(PROFILER_LATCHING_MODE_REGISTER, 0x40, servo=alias)
    mc.motion.motor_enable(servo=alias)
    pos_res = mc.configuration.get_position_feedback_resolution(servo=alias)
    init_pos = int(mean_actual_velocity_position(mc, alias))
    mc.motion.move_to_position(init_pos + pos_res, servo=alias, target_latch=False)
    test_act_pos = mean_actual_velocity_position(mc, alias)
    time.sleep(1)
    rel_tolerance = pos_res * POSITION_PERCENTAGE_ERROR_ALLOWED / 100
    assert pytest.approx(init_pos, rel_tolerance) == test_act_pos
    mc.motion.target_latch(servo=alias)
    time.sleep(1)
    test_act_pos = mean_actual_velocity_position(mc, alias)
    assert pytest.approx(init_pos + pos_res, rel_tolerance) == test_act_pos


@pytest.mark.virtual
@pytest.mark.smoke
@pytest.mark.parametrize("operation_mode", list(OperationMode))
def test_set_operation_mode(motion_controller, operation_mode):
    mc, alias = motion_controller
    mc.motion.set_operation_mode(operation_mode, servo=alias)
    test_op = mc.communication.get_register(OPERATION_MODE_REGISTER, servo=alias)
    assert operation_mode.value == test_op


@pytest.mark.smoke
@pytest.mark.parametrize("operation_mode", list(OperationMode))
def test_get_operation_mode(motion_controller, operation_mode):
    mc, alias = motion_controller
    mc.communication.set_register(OPERATION_MODE_REGISTER, operation_mode, servo=alias)
    test_op = mc.motion.get_operation_mode(servo=alias)
    assert test_op == operation_mode.value


@pytest.mark.smoke
def test_motor_enable(motion_controller):
    mc, alias = motion_controller
    mc.motion.motor_enable(servo=alias)
    assert mc.configuration.is_motor_enabled(servo=alias)


@pytest.mark.smoke
@pytest.mark.parametrize(
    "uid, value, exception_type, message",
    [
        ("DRV_PROT_USER_UNDER_VOLT", 100, exceptions.ILError, "User Under-voltage detected"),
        (
            "DRV_PROT_USER_OVER_TEMP",
            1,
            exceptions.ILError,
            "Over-temperature detected (user limit)",
        ),
        ("DRV_PROT_USER_OVER_VOLT", 1, exceptions.ILError, "User Over-voltage detected"),
    ],
)
def test_motor_enable_error(motion_controller_teardown, uid, value, exception_type, message):
    mc, alias = motion_controller_teardown
    mc.communication.set_register(uid, value, alias)
    with pytest.raises(exception_type) as excinfo:
        mc.motion.motor_enable(servo=alias)
    assert str(excinfo.value) == "An error occurred enabling motor. Reason: {}".format(message)


@pytest.mark.smoke
def test_motor_enable_with_fault(motion_controller_teardown):
    uid = "DRV_PROT_USER_UNDER_VOLT"
    value = 100
    exception_type = exceptions.ILError
    message = "User Under-voltage detected"
    mc, alias = motion_controller_teardown
    mc.communication.set_register(uid, value, alias)
    with pytest.raises(exception_type) as excinfo_1:
        mc.motion.motor_enable(servo=alias)
    assert str(excinfo_1.value) == "An error occurred enabling motor. Reason: {}".format(message)
    with pytest.raises(exception_type) as excinfo_2:
        mc.motion.motor_enable(servo=alias)
    assert str(excinfo_2.value) == "An error occurred enabling motor. Reason: {}".format(message)


@pytest.mark.smoke
@pytest.mark.parametrize("enable_motor", [True, False])
def test_motor_disable(motion_controller, enable_motor):
    mc, alias = motion_controller
    if enable_motor:
        mc.motion.motor_enable(servo=alias)
    mc.motion.motor_disable(servo=alias)
    assert not mc.configuration.is_motor_enabled(servo=alias)


@pytest.mark.smoke
def test_motor_disable_with_fault(motion_controller_teardown):
    uid = "DRV_PROT_USER_UNDER_VOLT"
    value = 100
    exception_type = exceptions.ILError
    mc, alias = motion_controller_teardown
    mc.communication.set_register(uid, value, alias)
    with pytest.raises(exception_type):
        mc.motion.motor_enable(servo=alias)
    mc.motion.motor_disable(servo=alias)
    assert not mc.configuration.is_motor_enabled(servo=alias)


@pytest.mark.smoke
def test_fault_reset(motion_controller_teardown):
    mc, alias = motion_controller_teardown
    uid = "DRV_PROT_USER_UNDER_VOLT"
    value = 100
    mc.communication.set_register(uid, value, alias)
    assert not mc.errors.is_fault_active(servo=alias)
    with pytest.raises(exceptions.ILError):
        mc.motion.motor_enable(servo=alias)
    assert mc.errors.is_fault_active(servo=alias)
    mc.motion.fault_reset(servo=alias)
    assert not mc.errors.is_fault_active(servo=alias)


@pytest.mark.virtual
@pytest.mark.smoke
@pytest.mark.parametrize("position_value", [1000, 0, -1000, 4000])
def test_set_position(motion_controller, position_value):
    mc, alias = motion_controller
    mc.motion.move_to_position(position_value, servo=alias, target_latch=False, blocking=False)
    test_position = mc.communication.get_register(POSITION_SET_POINT_REGISTER, servo=alias)
    assert test_position == position_value


@pytest.mark.smoke
@pytest.mark.parametrize("position_value", [1000, 0, -1000, 4000])
def test_move_position(motion_controller, position_value):
    mc, alias = motion_controller
    pos_res = mc.configuration.get_position_feedback_resolution(servo=alias)
    mc.motion.set_operation_mode(OperationMode.PROFILE_POSITION, servo=alias)
    mc.motion.motor_enable(servo=alias)
    mc.motion.move_to_position(position_value, servo=alias, blocking=True, timeout=10)
    test_position = mean_actual_velocity_position(mc, alias)
    pos_tolerance = pos_res * POSITION_PERCENTAGE_ERROR_ALLOWED / 100
    assert pytest.approx(position_value, abs=pos_tolerance) == test_position


@pytest.mark.virtual
@pytest.mark.smoke
@pytest.mark.parametrize("velocity_value", [0.5, 1, 0, -0.5])
def test_set_velocity(motion_controller, velocity_value):
    mc, alias = motion_controller
    mc.motion.set_velocity(velocity_value, servo=alias, target_latch=False)
    test_vel = mc.communication.get_register(VELOCITY_SET_POINT_REGISTER, servo=alias)
    assert test_vel == velocity_value


@pytest.mark.smoke
# TODO Update approx error. Well tuned motor is needed.
@pytest.mark.parametrize("velocity_value", [0.5, 1, 0, -0.5])
def test_set_velocity_blocking(motion_controller, velocity_value):
    mc, alias = motion_controller
    mc.motion.set_operation_mode(OperationMode.PROFILE_VELOCITY, servo=alias)
    mc.motion.motor_enable(servo=alias)
    mc.motion.set_velocity(velocity_value, servo=alias, blocking=True, timeout=10)
    time.sleep(1)
    test_vel = mean_actual_velocity_position(mc, alias, velocity=True)
    assert pytest.approx(velocity_value, abs=0.1) == test_vel


@pytest.mark.virtual
@pytest.mark.smoke
@pytest.mark.parametrize("current_value", [0.5, 1, 0, -0.5])
def test_set_current_quadrature(motion_controller, current_value):
    mc, alias = motion_controller
    mc.motion.set_current_quadrature(current_value, servo=alias)
    test_current = mc.communication.get_register(CURRENT_QUADRATURE_SET_POINT_REGISTER, servo=alias)
    assert pytest.approx(current_value) == test_current


@pytest.mark.virtual
@pytest.mark.smoke
@pytest.mark.parametrize("current_value", [0.5, 1, 0, -0.5])
def test_set_current_direct(motion_controller, current_value):
    mc, alias = motion_controller
    mc.motion.set_current_direct(current_value, servo=alias)
    test_current = mc.communication.get_register(CURRENT_DIRECT_SET_POINT_REGISTER, servo=alias)
    assert pytest.approx(current_value) == test_current


@pytest.mark.virtual
@pytest.mark.smoke
@pytest.mark.parametrize("voltage_value", [0.5, 1, 0, -0.5])
def test_set_voltage_quadrature(motion_controller, voltage_value):
    mc, alias = motion_controller
    mc.motion.set_voltage_quadrature(voltage_value, servo=alias)
    test_voltage = mc.communication.get_register(VOLTAGE_QUADRATURE_SET_POINT_REGISTER, servo=alias)
    assert pytest.approx(voltage_value) == test_voltage


@pytest.mark.virtual
@pytest.mark.smoke
@pytest.mark.parametrize("voltage_value", [0.5, 1, 0, -0.5])
def test_set_voltage_direct(motion_controller, voltage_value):
    mc, alias = motion_controller
    mc.motion.set_voltage_direct(voltage_value, servo=alias)
    test_voltage = mc.communication.get_register(VOLTAGE_DIRECT_SET_POINT_REGISTER, servo=alias)
    assert pytest.approx(voltage_value) == test_voltage


@pytest.mark.virtual
@pytest.mark.smoke
@pytest.mark.parametrize(
    "init_v, final_v, total_t, t, result",
    [
        (0, 1, 2, [1], [0.5]),
        (0, -2, 2, [1], [-1]),
        (-2, -4, 2, [1, 2], [-3, -4]),
        (-1, 1, 10, [2.5, 5, 7.5], [-0.5, 0, 0.5]),
        (0, 10, 10, [1, 2, 3], [1, 2, 3]),
        (0, 10, 10, [20], [10]),
        (-2.54, 23.45, 34, [2, 16.7, 44], [-1.01117647, 10.2256764, 23.45]),
    ],
)
def test_ramp_generator(mocker, init_v, final_v, total_t, t, result):
    mocker.patch("time.time", side_effect=[0, *t])
    generator = Motion.ramp_generator(init_v, final_v, total_t)
    first_val = next(generator)
    assert pytest.approx(init_v) == first_val
    for result_v in result:
        test_result = next(generator)
        assert pytest.approx(result_v) == test_result


@pytest.mark.smoke
@pytest.mark.parametrize("position_value", [-4000, -1000, 1000, 4000])
def test_get_actual_position(motion_controller, position_value):
    mc, alias = motion_controller
    mc.motion.set_operation_mode(OperationMode.PROFILE_POSITION, servo=alias)
    mc.motion.motor_enable(servo=alias)
    mc.motion.move_to_position(position_value, servo=alias, blocking=True, timeout=10)
    n_samples = 200
    test_position = np.zeros(n_samples)
    reg_value = np.zeros(n_samples)
    for sample_ix in range(n_samples):
        test_position[sample_ix] = mc.motion.get_actual_position(servo=alias)
        reg_value[sample_ix] = mc.communication.get_register(ACTUAL_POSITION_REGISTER, servo=alias)
    assert np.abs(np.mean(test_position) - np.mean(reg_value)) < 0.5


@pytest.mark.parametrize("velocity_value", [1, 0, -1])
def test_get_actual_velocity(motion_controller, velocity_value):
    mc, alias = motion_controller
    mc.motion.set_operation_mode(OperationMode.PROFILE_VELOCITY, servo=alias)
    mc.motion.motor_enable(servo=alias)
    mc.motion.set_velocity(velocity_value, servo=alias, blocking=True, timeout=10)
    time.sleep(2)
    n_samples = 200
    test_velocity = np.zeros(n_samples)
    reg_value = np.zeros(n_samples)
    for sample_ix in range(n_samples):
        test_velocity[sample_ix] = mc.motion.get_actual_velocity(servo=alias)
        reg_value[sample_ix] = mc.communication.get_register(ACTUAL_VELOCITY_REGISTER, servo=alias)
    assert np.abs(np.mean(test_velocity) - np.mean(reg_value)) < 0.1


@pytest.mark.virtual
@pytest.mark.smoke
def test_get_actual_current_direct(mocker, motion_controller):
    mc, alias = motion_controller
    patch_get_register = mocker.patch("ingeniamotion.communication.Communication.get_register")
    patch_get_register.return_value = 2.0
    mc.motion.get_actual_current_direct(servo=alias)
    patch_get_register.assert_called_once_with(ACTUAL_DIRECT_CURRENT_REGISTER, servo=alias, axis=1)


@pytest.mark.virtual
@pytest.mark.smoke
def test_get_actual_current_quadrature(mocker, motion_controller):
    mc, alias = motion_controller
    patch_get_register = mocker.patch("ingeniamotion.communication.Communication.get_register")
    patch_get_register.return_value = 2.0
    mc.motion.get_actual_current_quadrature(servo=alias)
    patch_get_register.assert_called_once_with(
        ACTUAL_QUADRATURE_CURRENT_REGISTER, servo=alias, axis=1
    )


@pytest.mark.parametrize(
    "function",
    [
        "wait_for_position",
        "wait_for_velocity",
    ],
)
@pytest.mark.virtual
def test_wait_for_function_timeout(motion_controller, function):
    timeout_value = 2
    mc, alias = motion_controller
    init_time = time.time()
    with pytest.raises(IMTimeoutError):
        getattr(mc.motion, function)(1000, servo=alias, timeout=timeout_value)
    final_time = time.time()
    assert pytest.approx(timeout_value, abs=0.1) == final_time - init_time


@pytest.mark.smoke
@pytest.mark.parametrize("op_mode", [OperationMode.VOLTAGE, OperationMode.CURRENT])
def test_set_internal_generator_configuration(motion_controller_teardown, op_mode):
    mc, alias = motion_controller_teardown
    mc.motion.set_internal_generator_configuration(op_mode, servo=alias)
    assert op_mode == mc.motion.get_operation_mode(servo=alias)
    assert 1 == mc.configuration.get_motor_pair_poles(servo=alias)


# TODO: Remove skip after fixing INGM-349
@pytest.mark.parametrize("op_mode", [OperationMode.VOLTAGE, OperationMode.CURRENT])
@pytest.mark.parametrize("direction", [-1, 1])
@pytest.mark.skip
def test_internal_generator_saw_tooth_move(motion_controller_teardown, op_mode, direction):
    mc, alias = motion_controller_teardown
    pair_poles = mc.configuration.get_motor_pair_poles(servo=alias)
    pos_resolution = mc.configuration.get_position_feedback_resolution(servo=alias)
    mc.motion.set_internal_generator_configuration(op_mode, servo=alias)
    mc.motion.motor_enable(servo=alias)
    if op_mode == OperationMode.CURRENT:
        mc.motion.current_quadrature_ramp(1, 1, servo=alias)
        mc.motion.current_direct_ramp(1, 1, servo=alias)
    else:
        mc.motion.voltage_quadrature_ramp(1, 1, servo=alias)
        mc.motion.voltage_direct_ramp(1, 1, servo=alias)
    time.sleep(1)
    initial_position = mc.motion.get_actual_position(servo=alias)
    mc.motion.internal_generator_saw_tooth_move(direction, pair_poles, 1, servo=alias)
    time.sleep(pair_poles)
    time.sleep(1)
    final_position = mc.motion.get_actual_position(servo=alias)
    total_movement = final_position - initial_position
    expected_movement = pos_resolution * direction
    assert (
        abs(total_movement - expected_movement)
        < pos_resolution * POSITION_PERCENTAGE_ERROR_ALLOWED / 100
    )


# TODO: Remove skip after fixing INGM-349
@pytest.mark.parametrize("op_mode", [OperationMode.VOLTAGE, OperationMode.CURRENT])
@pytest.mark.parametrize("direction", [-1, 1])
@pytest.mark.skip
def test_internal_generator_constant_move(motion_controller_teardown, op_mode, direction):
    mc, alias = motion_controller_teardown
    pair_poles = mc.configuration.get_motor_pair_poles(servo=alias)
    pos_resolution = mc.configuration.get_position_feedback_resolution(servo=alias)
    cycle_pos = pos_resolution / pair_poles
    mc.motion.set_internal_generator_configuration(op_mode, servo=alias)
    mc.motion.motor_enable(servo=alias)
    if op_mode == OperationMode.CURRENT:
        mc.motion.current_quadrature_ramp(1, 2, servo=alias)
        mc.motion.current_direct_ramp(1, 2, servo=alias)
    else:
        mc.motion.voltage_quadrature_ramp(1, 2, servo=alias)
        mc.motion.voltage_direct_ramp(1, 2, servo=alias)
    time.sleep(1)
    mc.motion.internal_generator_constant_move(0, servo=alias)
    initial_position = mc.motion.get_actual_position(servo=alias)
    list_values = np.linspace(0, 1, 5) if direction > 0 else np.linspace(1, 0, 5)
    for value in list_values:
        mc.motion.internal_generator_constant_move(value, servo=alias)
        time.sleep(1)
        final_position = mc.motion.get_actual_position(servo=alias)
        total_movement = final_position - initial_position
        expected_movement = cycle_pos * value if direction > 0 else cycle_pos * (value - 1)
        assert (
            abs(total_movement - expected_movement)
            < pos_resolution * POSITION_PERCENTAGE_ERROR_ALLOWED / 100
        )


@pytest.mark.parametrize(
    "function",
    [
        "target_latch",
        "get_operation_mode",
        "get_actual_position",
        "get_actual_velocity",
        "get_actual_current_direct",
        "get_actual_current_quadrature",
    ],
)
@pytest.mark.virtual
def test_wrong_type_exception(mocker, motion_controller, function):
    mc, alias = motion_controller
    mocker.patch.object(mc.communication, "get_register", return_value="invalid_value")
    with pytest.raises(TypeError):
        getattr(mc.motion, function)(servo=alias)


@pytest.mark.virtual
def test_set_internal_generator_configuration_exception(motion_controller):
    mc, alias = motion_controller
    with pytest.raises(ValueError):
        mc.motion.set_internal_generator_configuration(OperationMode.VELOCITY, servo=alias)



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

import ingenialogger
from ingenialink.exceptions import ILError

from ingeniamotion.exceptions import IMRegisterNotExist, IMRegisterWrongAccess
from ingeniamotion.metaclass import DEFAULT_SERVO
from ingeniamotion.wizard_tests.stoppable import StopException, Stoppable

if TYPE_CHECKING:
    from ingeniamotion import MotionController

from ingeniamotion.enums import SeverityLevel


class TestError(Exception):
    pass


class BaseTest(ABC, Stoppable):
    WARNING_BIT_MASK = 0x0FFFFFFF

    def __init__(self) -> None:
        self.backup_registers_names: List[str] = []
        self.optional_backup_registers_names: List[str] = []
        self.backup_registers: Dict[int, Dict[str, Union[int, float, str]]] = {}
        self.suggested_registers: Dict[str, Union[int, float, str]] = {}
        self.mc: "MotionController"
        self.servo: str = DEFAULT_SERVO
        self.axis: int = 0
        self.report: Optional[Dict[str, Any]] = None
        self.logger = ingenialogger.get_logger(__name__)

    def save_backup_registers(self) -> None:
        self.backup_registers[self.axis] = {}
        for uid in self.backup_registers_names:
            try:
                value = self.mc.communication.get_register(uid, servo=self.servo, axis=self.axis)
                self.backup_registers[self.axis][uid] = value
            except IMRegisterNotExist as e:
                self.logger.warning(e, axis=self.axis)

        for uid in self.optional_backup_registers_names:
            if self.mc.info.register_exists(uid, self.axis, self.servo):
                value = self.mc.communication.get_register(uid, servo=self.servo, axis=self.axis)
                self.backup_registers[self.axis][uid] = value

    def restore_backup_registers(self) -> None:
        """Restores the value of the registers after the test execution.

        Notes:
        This should only be called by the Wizard.
        """
        for subnode, registers in self.backup_registers.items():
            for key, value in self.backup_registers[subnode].items():
                try:
                    self.mc.communication.set_register(key, value, servo=self.servo, axis=self.axis)
                except IMRegisterNotExist as e:
                    self.logger.warning(e, axis=subnode)
                except IMRegisterWrongAccess as e:
                    self.logger.warning(e, axis=subnode)

    @Stoppable.stoppable
    def show_error_message(self) -> None:
        error_code, axis, warning = self.mc.errors.get_last_buffer_error(
            servo=self.servo, axis=self.axis
        )
        *_, error_msg = self.mc.errors.get_error_data(error_code, servo=self.servo)
        raise TestError(error_msg)

    @abstractmethod
    def setup(self) -> None:
        pass

    @abstractmethod
    def loop(self) -> Any:
        pass

    @abstractmethod
    def teardown(self) -> None:
        pass

    def run(
        self,
    ) -> Optional[Dict[str, Union[SeverityLevel, Dict[str, Union[int, float, str]], str]]]:
        self.reset_stop()
        self.save_backup_registers()
        try:
            self.setup()
            output = self.loop()
            self.report = self.generate_report(output)
        except ILError as err:
            raise err
        except StopException:
            self.logger.warning("Test has been stopped")
        finally:
            try:
                self.teardown()
            finally:
                self.restore_backup_registers()
        return self.report

    def generate_report(
        self, output: Any
    ) -> Dict[str, Union[SeverityLevel, Dict[str, Union[int, float, str]], str]]:
        return {
            "result_severity": self.get_result_severity(output),
            "suggested_registers": self.suggested_registers,
            "result_message": self.get_result_msg(output),
        }

    @abstractmethod
    def get_result_msg(self, output: Any) -> str:
        pass

    @abstractmethod
    def get_result_severity(self, output: Any) -> SeverityLevel:
        pass



import time
from threading import Thread

import pytest
from ingenialink import exceptions

from ingeniamotion.enums import SensorType, SeverityLevel
from ingeniamotion.exceptions import IMRegisterNotExist
from ingeniamotion.wizard_tests.base_test import TestError
from ingeniamotion.wizard_tests.feedbacks_tests.absolute_encoder1_test import AbsoluteEncoder1Test
from ingeniamotion.wizard_tests.feedbacks_tests.absolute_encoder2_test import AbsoluteEncoder2Test
from ingeniamotion.wizard_tests.feedbacks_tests.digital_hall_test import DigitalHallTest
from ingeniamotion.wizard_tests.feedbacks_tests.digital_incremental1_test import (
    DigitalIncremental1Test,
)
from ingeniamotion.wizard_tests.feedbacks_tests.digital_incremental2_test import (
    DigitalIncremental2Test,
)
from ingeniamotion.wizard_tests.feedbacks_tests.secondary_ssi_test import SecondarySSITest
from ingeniamotion.wizard_tests.phase_calibration import Phasing
from ingeniamotion.wizard_tests.phasing_check import PhasingCheck

CURRENT_QUADRATURE_SET_POINT_REGISTER = "CL_CUR_Q_SET_POINT"
RATED_CURRENT_REGISTER = "MOT_RATED_CURRENT"
MAXIMUM_CONTINUOUS_CURRENT_DRIVE_PROTECTION = "DRV_PROT_MAN_MAX_CONT_CURRENT_VALUE"


@pytest.fixture
def force_fault(motion_controller):
    mc, alias = motion_controller
    uid = "DRV_PROT_USER_UNDER_VOLT"
    mc.communication.set_register(uid, 100, alias)
    yield exceptions.ILError
    mc.communication.set_register(uid, 10, alias)


@pytest.fixture(scope="module")
def feedback_test_setup(motion_controller):
    mc, alias = motion_controller
    mc.tests.commutation(servo=alias)


@pytest.mark.usefixtures("feedback_test_setup")
def test_digital_halls_test(motion_controller, feedback_list):
    mc, alias = motion_controller
    commutation_fdbk = mc.configuration.get_commutation_feedback(servo=alias)
    if SensorType.HALLS in feedback_list:
        results = mc.tests.digital_halls_test(servo=alias)
        assert results["result_severity"] == SeverityLevel.SUCCESS
    else:
        with pytest.raises(TestError):
            mc.tests.digital_halls_test(servo=alias)
    assert commutation_fdbk == mc.configuration.get_commutation_feedback(servo=alias)


@pytest.mark.usefixtures("feedback_test_setup")
def test_incremental_encoder_1_test(motion_controller, feedback_list):
    mc, alias = motion_controller
    commutation_fdbk = mc.configuration.get_commutation_feedback(servo=alias)
    if SensorType.QEI in feedback_list:
        results = mc.tests.incremental_encoder_1_test(servo=alias)
        assert results["result_severity"] == SeverityLevel.SUCCESS
    else:
        with pytest.raises(TestError):
            mc.tests.incremental_encoder_1_test(servo=alias)
    assert commutation_fdbk == mc.configuration.get_commutation_feedback(servo=alias)


@pytest.mark.usefixtures("feedback_test_setup")
def test_incremental_encoder_2_test(motion_controller, feedback_list):
    mc, alias = motion_controller
    if not mc.info.register_exists("FBK_DIGENC2_RESOLUTION", servo=alias):
        pytest.skip("Incremental encoder 2 is not available")
    commutation_fdbk = mc.configuration.get_commutation_feedback(servo=alias)
    if SensorType.QEI2 in feedback_list:
        results = mc.tests.incremental_encoder_2_test(servo=alias)
        assert results["result_severity"] == SeverityLevel.SUCCESS
    else:
        with pytest.raises(TestError):
            mc.tests.incremental_encoder_2_test(servo=alias)
    assert commutation_fdbk == mc.configuration.get_commutation_feedback(servo=alias)


@pytest.mark.usefixtures("feedback_test_setup")
def test_absolute_encoder_1_test(motion_controller, feedback_list):
    mc, alias = motion_controller
    commutation_fdbk = mc.configuration.get_commutation_feedback(servo=alias)
    if SensorType.ABS1 in feedback_list:
        results = mc.tests.absolute_encoder_1_test(servo=alias)
        assert results["result_severity"] == SeverityLevel.SUCCESS
    else:
        with pytest.raises(TestError):
            mc.tests.absolute_encoder_1_test(servo=alias)
    assert commutation_fdbk == mc.configuration.get_commutation_feedback(servo=alias)


@pytest.mark.usefixtures("feedback_test_setup")
def test_absolute_encoder_2_test(motion_controller, feedback_list):
    mc, alias = motion_controller
    commutation_fdbk = mc.configuration.get_commutation_feedback(servo=alias)
    if SensorType.BISSC2 in feedback_list:
        results = mc.tests.absolute_encoder_2_test(servo=alias)
        assert results["result_severity"] == SeverityLevel.SUCCESS
    else:
        with pytest.raises(TestError):
            mc.tests.absolute_encoder_2_test(servo=alias)
    assert commutation_fdbk == mc.configuration.get_commutation_feedback(servo=alias)


@pytest.mark.usefixtures("feedback_test_setup")
def test_secondary_ssi_test(motion_controller, feedback_list):
    mc, alias = motion_controller
    commutation_fdbk = mc.configuration.get_commutation_feedback(servo=alias)
    if SensorType.QEI in feedback_list:
        pytest.skip("Can not run the test. Incremental encoder 1 and SSI 2 share pins.")
    if SensorType.SSI2 in feedback_list:
        results = mc.tests.secondary_ssi_test(servo=alias)
        assert results["result_severity"] == SeverityLevel.SUCCESS
    else:
        with pytest.raises(TestError):
            mc.tests.secondary_ssi_test(servo=alias)
    assert commutation_fdbk == mc.configuration.get_commutation_feedback(servo=alias)


def test_commutation(motion_controller_teardown):
    mc, alias = motion_controller_teardown
    results = mc.tests.commutation(servo=alias)
    assert results["result_severity"] == SeverityLevel.SUCCESS


@pytest.mark.smoke
def test_commutation_error(motion_controller, force_fault):
    mc, alias = motion_controller
    with pytest.raises(force_fault):
        mc.tests.commutation(servo=alias)


@pytest.mark.skip("Skip until is fixed INGM-352")
def test_phasing_check(motion_controller):
    mc, alias = motion_controller
    mc.tests.commutation(servo=alias)
    results = mc.tests.phasing_check(servo=alias)
    assert results["result_severity"] == SeverityLevel.SUCCESS


@pytest.mark.smoke
def test_phasing_check_error(motion_controller, force_fault):
    mc, alias = motion_controller
    with pytest.raises(force_fault):
        mc.tests.phasing_check(servo=alias)


@pytest.mark.smoke
def test_sto_test(motion_controller):
    mc, alias = motion_controller
    results = mc.tests.sto_test(servo=alias)
    assert results["result_severity"] == SeverityLevel.SUCCESS


@pytest.mark.virtual
@pytest.mark.smoke
@pytest.mark.parametrize(
    "sto_value, message",
    [
        (0x4, "STO Active"),
        (0x1F, "Abnormal STO Latched"),
        (0x8, "Abnormal STO"),
        (0x73, "Abnormal Supply"),
        (0x5, "STO Inputs Differ"),
    ],
)
def test_sto_test_error(mocker, motion_controller, sto_value, message):
    mocker.patch("ingeniamotion.configuration.Configuration.get_sto_status", return_value=sto_value)
    mc, alias = motion_controller
    results = mc.tests.sto_test(servo=alias)
    assert results["result_severity"] == SeverityLevel.FAIL
    assert results["result_message"] == message


@pytest.mark.smoke
def test_brake_test(motion_controller):
    mc, alias = motion_controller
    pair_poles = mc.configuration.get_motor_pair_poles(servo=alias)
    brake_test = mc.tests.brake_test(servo=alias)
    assert 1 == mc.configuration.get_motor_pair_poles(servo=alias)
    brake_test.finish()
    assert pair_poles == mc.configuration.get_motor_pair_poles(servo=alias)


def get_backup_registers(test, mc, alias):
    reg_values = {}
    for reg in test.BACKUP_REGISTERS:
        try:
            reg_values[reg] = mc.communication.get_register(reg, servo=alias)
        except IMRegisterNotExist:
            pass
    return reg_values


def run_test_and_stop(test):
    test_thread = Thread(target=test.run)
    test_thread.start()
    time.sleep(2)
    test.stop()
    test_thread.join()


@pytest.mark.usefixtures("feedback_test_setup")
@pytest.mark.parametrize(
    "feedback_class",
    [
        AbsoluteEncoder1Test,
        AbsoluteEncoder2Test,
        DigitalHallTest,
        DigitalIncremental1Test,
        DigitalIncremental2Test,
        SecondarySSITest,
    ],
)
def test_feedback_stop(motion_controller, feedback_class):
    mc, alias = motion_controller
    test = feedback_class(mc, alias, 1)
    reg_values = get_backup_registers(test, mc, alias)
    run_test_and_stop(test)
    for reg in reg_values:
        assert reg_values[reg] == mc.communication.get_register(reg, servo=alias)


@pytest.mark.virtual
def test_commutation_stop(motion_controller):
    mc, alias = motion_controller
    test = Phasing(mc, alias, 1)
    reg_values = get_backup_registers(test, mc, alias)
    run_test_and_stop(test)
    for reg in reg_values:
        assert reg_values[reg] == mc.communication.get_register(reg, servo=alias)


@pytest.mark.virtual
def test_phasing_check_stop(motion_controller):
    mc, alias = motion_controller
    test = PhasingCheck(mc, alias, 1)
    reg_values = get_backup_registers(test, mc, alias)
    run_test_and_stop(test)
    for reg in reg_values:
        assert reg_values[reg] == mc.communication.get_register(reg, servo=alias)


@pytest.mark.virtual
@pytest.mark.parametrize("test_currents", ["Rated current", "Drive current", "Same value"])
@pytest.mark.parametrize(
    "test_sensor",
    [
        SensorType.ABS1,
        SensorType.QEI,
        SensorType.HALLS,
        SensorType.SSI2,
        SensorType.BISSC2,
        SensorType.QEI2,
    ],
)
def test_current_ramp_up(motion_controller, test_currents, test_sensor):
    mc, alias = motion_controller

    axis = 1
    test_feedback_options = {
        SensorType.ABS1: AbsoluteEncoder1Test(mc, alias, axis),
        SensorType.QEI: DigitalIncremental1Test(mc, alias, axis),
        SensorType.HALLS: DigitalHallTest(mc, alias, axis),
        SensorType.SSI2: SecondarySSITest(mc, alias, axis),
        SensorType.BISSC2: AbsoluteEncoder2Test(mc, alias, axis),
        SensorType.QEI2: DigitalIncremental2Test(mc, alias, axis),
    }
    feedbacks_test = test_feedback_options[test_sensor]

    current_drive = mc.communication.get_register(
        MAXIMUM_CONTINUOUS_CURRENT_DRIVE_PROTECTION, servo=alias, axis=1
    )

    if test_currents == "Rated current":
        current_motor = current_drive + 1
    elif test_currents == "Drive current":
        current_motor = current_drive - 1
    else:
        current_motor = current_drive

    mc.communication.set_register(RATED_CURRENT_REGISTER, current_motor, servo=alias, axis=1)

    feedbacks_test.current_ramp_up()

    current_quadrature = mc.communication.get_register(
        CURRENT_QUADRATURE_SET_POINT_REGISTER, servo=alias, axis=1
    )

    test_max_current = current_quadrature / feedbacks_test.PERCENTAGE_CURRENT_USED

    if test_currents == "Rated current":
        assert pytest.approx(test_max_current) == current_drive
    elif test_currents == "Drive current":
        assert pytest.approx(test_max_current) == current_motor
    else:
        assert pytest.approx(test_max_current) == current_drive == current_motor



class IMException(Exception):
    """Exceptions raised by IngeniaMotion"""

    pass


class IMMonitoringError(IMException):
    """Monitoring error raised by IngeniaMotion"""

    pass


class IMDisturbanceError(IMException):
    """Disturbance error raised by IngeniaMotion"""

    pass


class IMStatusWordError(IMException):
    """Status word error raised by IngeniaMotion"""

    pass


class IMRegisterNotExist(IMException):
    """Error raised by IngeniaMotion when a register not exists"""

    pass


class IMRegisterWrongAccess(IMException):
    """Error raised by IngeniaMotion when trying to write to a read-only register"""

    pass


class IMTimeoutError(IMException):
    """Error raised by IngeniaMotion when a timeout has occurred"""

    pass



import json
import os
import time
from typing import Dict

import numpy as np
import pytest
from virtual_drive.core import VirtualDrive

from ingeniamotion import MotionController
from ingeniamotion.enums import CAN_BAUDRATE, CAN_DEVICE, SensorType

ALLOW_PROTOCOLS = ["eoe", "soem", "canopen", "virtual"]

test_report_key = pytest.StashKey[Dict[str, pytest.CollectReport]]()


def pytest_addoption(parser):
    parser.addoption(
        "--protocol", action="store", default="eoe", help="eoe, soem", choices=ALLOW_PROTOCOLS
    )
    parser.addoption("--slave", type="int", default=0, help="Slave index in config.json")


def pytest_collection_modifyitems(config, items):
    protocol = config.getoption("--protocol")
    negate_protocols = [x for x in ALLOW_PROTOCOLS if x != protocol]
    skip_by_protocol = pytest.mark.skip(reason="Protocol does not match")
    for item in items:
        if protocol in item.keywords:
            continue
        for not_protocol in negate_protocols:
            if not_protocol in item.keywords:
                item.add_marker(skip_by_protocol)


@pytest.fixture(scope="session")
def read_config(request):
    slave = request.config.getoption("--slave")
    protocol = request.config.getoption("--protocol")
    config = "tests/config.json"
    with open(config, "r") as fp:
        contents = json.load(fp)
    relative_path = contents["virtual"][0]["dictionary"]
    absolute_path = os.path.join(os.path.abspath(os.getcwd()), relative_path)
    contents["virtual"][0]["dictionary"] = absolute_path
    return contents[protocol][slave]


def connect_eoe(mc, config, alias):
    mc.communication.connect_servo_eoe(config["ip"], config["dictionary"], alias=alias)


def connect_soem(mc, config, alias):
    mc.communication.connect_servo_ethercat_interface_index(
        config["index"],
        config["slave"],
        config["dictionary"],
        alias,
    )


def connect_canopen(mc, config, alias):
    device = CAN_DEVICE(config["device"])
    baudrate = CAN_BAUDRATE(config["baudrate"])
    mc.communication.connect_servo_canopen(
        device,
        config["dictionary"],
        config["node_id"],
        baudrate,
        config["channel"],
        alias=alias,
    )


@pytest.fixture(scope="session")
def motion_controller(pytestconfig, read_config):
    alias = "test"
    mc = MotionController()
    protocol = pytestconfig.getoption("--protocol")
    if protocol == "soem":
        connect_soem(mc, read_config, alias)
    elif protocol == "canopen":
        connect_canopen(mc, read_config, alias)
    elif protocol == "virtual":
        virtual_drive = VirtualDrive(read_config["port"], read_config["dictionary"])
        virtual_drive.start()
        connect_eoe(mc, read_config, alias)
    else:
        connect_eoe(mc, read_config, alias)

    if protocol != "virtual":
        mc.configuration.restore_configuration(servo=alias)
        mc.configuration.load_configuration(read_config["config_file"], servo=alias)
        yield mc, alias
        mc.communication.disconnect(alias)
    else:
        yield mc, alias
        virtual_drive.stop()


@pytest.fixture(autouse=True)
def disable_motor_fixture(pytestconfig, motion_controller):
    yield
    protocol = pytestconfig.getoption("--protocol")
    if protocol != "virtual":
        mc, alias = motion_controller
        mc.motion.motor_disable(servo=alias)
        mc.motion.fault_reset(servo=alias)


@pytest.fixture
def motion_controller_teardown(motion_controller, pytestconfig, read_config):
    yield motion_controller
    protocol = pytestconfig.getoption("--protocol")
    if protocol == "virtual":
        return
    mc, alias = motion_controller
    mc.motion.motor_disable(servo=alias)
    mc.configuration.load_configuration(read_config["config_file"], servo=alias)
    mc.motion.fault_reset(servo=alias)


@pytest.fixture
def disable_monitoring_disturbance(motion_controller):
    yield
    mc, alias = motion_controller
    mc.capture.clean_monitoring_disturbance(servo=alias)


@pytest.fixture(scope="session")
def feedback_list(motion_controller):
    mc, alias = motion_controller
    fdbk_lst = [
        mc.configuration.get_commutation_feedback(servo=alias),
        mc.configuration.get_reference_feedback(servo=alias),
        mc.configuration.get_velocity_feedback(servo=alias),
        mc.configuration.get_position_feedback(servo=alias),
        mc.configuration.get_auxiliar_feedback(servo=alias),
    ]
    return set(fdbk_lst)


@pytest.fixture
def clean_and_restore_feedbacks(motion_controller):
    mc, alias = motion_controller
    comm = mc.configuration.get_commutation_feedback(servo=alias)
    ref = mc.configuration.get_reference_feedback(servo=alias)
    vel = mc.configuration.get_velocity_feedback(servo=alias)
    pos = mc.configuration.get_position_feedback(servo=alias)
    aux = mc.configuration.get_auxiliar_feedback(servo=alias)
    mc.configuration.set_commutation_feedback(SensorType.INTGEN, servo=alias)
    mc.configuration.set_reference_feedback(SensorType.INTGEN, servo=alias)
    mc.configuration.set_velocity_feedback(SensorType.INTGEN, servo=alias)
    mc.configuration.set_position_feedback(SensorType.INTGEN, servo=alias)
    mc.configuration.set_auxiliar_feedback(SensorType.QEI, servo=alias)
    yield
    mc.configuration.set_commutation_feedback(comm, servo=alias)
    mc.configuration.set_reference_feedback(ref, servo=alias)
    mc.configuration.set_velocity_feedback(vel, servo=alias)
    mc.configuration.set_position_feedback(pos, servo=alias)
    mc.configuration.set_auxiliar_feedback(aux, servo=alias)


@pytest.fixture()
def skip_if_monitoring_not_available(motion_controller):
    mc, alias = motion_controller
    try:
        mc.capture._check_version(alias)
    except NotImplementedError:
        pytest.skip("Monitoring is not available")


@pytest.hookimpl(tryfirst=True, hookwrapper=True)
def pytest_runtest_makereport(item, call):
    # execute all other hooks to obtain the report object
    outcome = yield
    rep = outcome.get_result()

    # store test results for each phase of a call, which can be "setup", "call", "teardown"
    item.stash.setdefault(test_report_key, {})[rep.when] = rep


@pytest.fixture(scope="function", autouse=True)
def load_configuration_if_test_fails(pytestconfig, request, motion_controller, read_config):
    mc, alias = motion_controller
    yield

    report = request.node.stash[test_report_key]
    protocol = pytestconfig.getoption("--protocol")
    if protocol != "virtual" and (
        report["setup"].failed or ("call" not in report) or report["call"].failed
    ):
        mc.configuration.load_configuration(read_config["config_file"], servo=alias)
        mc.motion.fault_reset(servo=alias)


def mean_actual_velocity_position(mc, servo, velocity=False, n_samples=200, sampling_period=0):
    samples = np.zeros(n_samples)
    get_actual_value_dict = {
        True: mc.motion.get_actual_velocity,
        False: mc.motion.get_actual_position,
    }
    for sample_idx in range(n_samples):
        value = get_actual_value_dict[velocity](servo=servo)
        samples[sample_idx] = value
        time.sleep(sampling_period)
    return np.mean(samples)


@pytest.fixture(scope="module", autouse=True)
def load_configuration_after_each_module(pytestconfig, motion_controller, read_config):
    yield motion_controller
    protocol = pytestconfig.getoption("--protocol")
    if protocol != "virtual":
        mc, alias = motion_controller
        mc.motion.motor_disable(servo=alias)
        mc.configuration.load_configuration(read_config["config_file"], servo=alias)



import time

import pytest

from ingeniamotion.enums import HomingMode, SensorType, OperationMode
from tests.conftest import mean_actual_velocity_position

HOMING_MODE_REGISTER = "HOM_MODE"
HOMING_OFFSET_REGISTER = "HOM_OFFSET"
HOMING_TIMEOUT_REGISTER = "HOM_SEQ_TIMEOUT"
HOMING_ZERO_VELOCITY_REGISTER = "HOM_SPEED_ZERO"
HOMING_SEARCH_VELOCITY_REGISTER = "HOM_SPEED_SEARCH"
HOMING_INDEX_PULSE_SOURCE_REGISTER = "HOM_IDX_PULSE_SRC"
POSITIVE_HOMING_SWITCH_REGISTER = "IO_IN_POS_HOM_SWITCH"
NEGATIVE_HOMING_SWITCH_REGISTER = "IO_IN_NEG_HOM_SWITCH"
VELOCITY_SET_POINT_REGISTER = "CL_VEL_SET_POINT_VALUE"

STATUS_WORD_HOMING_ERROR_BIT = 0x2000
STATUS_WORD_HOMING_ATTAINED_BIT = 0x1000
STATUS_WORD_TARGET_REACHED_BIT = 0x400

RELATIVE_ERROR_ALLOWED = 3e-2


@pytest.fixture
def initial_position(motion_controller):
    mc, alias = motion_controller
    mc.motion.set_operation_mode(OperationMode.PROFILE_POSITION, servo=alias)
    mc.motion.motor_enable(servo=alias)
    last_pos = mc.motion.get_actual_position(servo=alias)
    position = mc.configuration.get_position_feedback_resolution(servo=alias) // 2
    mc.motion.move_to_position(position + last_pos, servo=alias, blocking=True, timeout=5)
    mc.motion.motor_disable(servo=alias)
    return position


@pytest.mark.virtual
@pytest.mark.smoke
@pytest.mark.parametrize("homing_mode", list(HomingMode))
def test_set_homing_mode(motion_controller, homing_mode):
    mc, alias = motion_controller
    mc.configuration.set_homing_mode(homing_mode, servo=alias)
    test_homing_mode = mc.communication.get_register(HOMING_MODE_REGISTER, servo=alias)
    assert test_homing_mode == homing_mode


@pytest.mark.virtual
@pytest.mark.smoke
@pytest.mark.parametrize("homing_offset", [0, 10, 500, -12, -100, 1000])
def test_set_homing_offset(motion_controller, homing_offset):
    mc, alias = motion_controller
    mc.configuration.set_homing_offset(homing_offset, servo=alias)
    test_homing_offset = mc.communication.get_register(HOMING_OFFSET_REGISTER, servo=alias)
    assert test_homing_offset == homing_offset


@pytest.mark.virtual
@pytest.mark.smoke
@pytest.mark.parametrize("homing_timeout", [0, 10, 500, 1000, 5000, 10000])
def test_set_homing_timeout(motion_controller, homing_timeout):
    mc, alias = motion_controller
    mc.configuration.set_homing_timeout(homing_timeout, servo=alias)
    test_homing_timeout = mc.communication.get_register(HOMING_TIMEOUT_REGISTER, servo=alias)
    assert test_homing_timeout == homing_timeout


@pytest.mark.parametrize("homing_offset", [0, 1000])
@pytest.mark.usefixtures("initial_position")
def test_homing_on_current_position(motion_controller, homing_offset):
    mc, alias = motion_controller
    mc.configuration.homing_on_current_position(homing_offset, servo=alias)
    feedback_resolution = mc.configuration.get_position_feedback_resolution(servo=alias)
    assert pytest.approx(
        homing_offset,
        abs=feedback_resolution * RELATIVE_ERROR_ALLOWED,
    ) == mc.motion.get_actual_position(servo=alias)


@pytest.mark.usefixtures("initial_position")
@pytest.mark.parametrize("direction", [1, 0])
def test_homing_on_switch_limit(motion_controller, direction):
    mc, alias = motion_controller
    homing_offset = 10
    homing_timeout = 5000
    search_vel = 10.0
    zero_vel = 1.0
    switch = 2
    mc.configuration.homing_on_switch_limit(
        homing_offset,
        direction,
        switch,
        homing_timeout,
        search_vel,
        zero_vel,
        servo=alias,
        motor_enable=False,
    )
    test_offset = mc.communication.get_register(HOMING_OFFSET_REGISTER, servo=alias)
    test_timeout = mc.communication.get_register(HOMING_TIMEOUT_REGISTER, servo=alias)
    test_hom_mode = mc.communication.get_register(HOMING_MODE_REGISTER, servo=alias)
    test_op_mode = mc.motion.get_operation_mode(servo=alias)
    test_search_vel = mc.communication.get_register(HOMING_SEARCH_VELOCITY_REGISTER, servo=alias)
    test_zero_vel = mc.communication.get_register(HOMING_ZERO_VELOCITY_REGISTER, servo=alias)
    switch_register = (
        POSITIVE_HOMING_SWITCH_REGISTER if direction == 1 else NEGATIVE_HOMING_SWITCH_REGISTER
    )
    test_switch = mc.communication.get_register(switch_register, servo=alias)
    assert test_offset == homing_offset
    assert test_timeout == homing_timeout
    if direction == 1:
        assert test_hom_mode == HomingMode.POSITIVE_LIMIT_SWITCH
    elif direction == 0:
        assert test_hom_mode == HomingMode.NEGATIVE_LIMIT_SWITCH
    assert test_op_mode == OperationMode.HOMING
    assert pytest.approx(zero_vel) == test_zero_vel
    assert pytest.approx(search_vel) == test_search_vel
    assert test_switch == switch


@pytest.mark.usefixtures("initial_position")
def test_homing_on_switch_limit_timeout(motion_controller):
    mc, alias = motion_controller
    homing_offset = 10
    homing_timeout = 5000
    search_vel = 10.0
    zero_vel = 1.0
    switch = 2
    direction = 1
    mc.configuration.homing_on_switch_limit(
        homing_offset,
        direction,
        switch,
        homing_timeout,
        search_vel,
        zero_vel,
        servo=alias,
        motor_enable=False,
    )
    time.sleep(homing_timeout / 1000)
    assert pytest.approx(0, abs=0.05) == mean_actual_velocity_position(mc, alias, velocity=True)
    mc.motion.motor_enable(servo=alias)
    mc.motion.target_latch(servo=alias)
    time.sleep(1)
    assert abs(mean_actual_velocity_position(mc, alias, velocity=True)) > 0.05
    time.sleep(homing_timeout / 1000)
    assert pytest.approx(0, abs=0.05) == mean_actual_velocity_position(mc, alias, velocity=True)


def __check_index_pulse_is_allowed(feedback_list):
    motor_enable = True
    if SensorType.QEI in feedback_list:
        sensor_index = 0
    elif SensorType.QEI2 in feedback_list:
        sensor_index = 1
    else:
        sensor_index = 1
        motor_enable = False
    return motor_enable, sensor_index


def __check_homing_was_successful(mc, alias, timeout_ms):
    init_time = time.time()
    while init_time + timeout_ms / 1000 > time.time():
        status_word = mc.configuration.get_status_word(servo=alias)
        homing_error = bool(status_word & STATUS_WORD_HOMING_ERROR_BIT)
        homing_attained = bool(status_word & STATUS_WORD_HOMING_ATTAINED_BIT)
        if (not homing_error) & homing_attained:
            return True
    return False


@pytest.mark.usefixtures("initial_position")
@pytest.mark.parametrize("direction", [1, 0])
def test_homing_on_index_pulse(motion_controller, feedback_list, direction):
    mc, alias = motion_controller
    homing_offset = 1000
    homing_timeout = 10000
    zero_vel = 0.1
    motor_enable, sensor_index = __check_index_pulse_is_allowed(feedback_list)
    mc.configuration.homing_on_index_pulse(
        homing_offset,
        direction,
        sensor_index,
        homing_timeout,
        zero_vel,
        servo=alias,
        motor_enable=motor_enable,
    )
    if motor_enable:
        assert __check_homing_was_successful(mc, alias, homing_timeout)
    test_offset = mc.communication.get_register(HOMING_OFFSET_REGISTER, servo=alias)
    test_timeout = mc.communication.get_register(HOMING_TIMEOUT_REGISTER, servo=alias)
    test_hom_mode = mc.communication.get_register(HOMING_MODE_REGISTER, servo=alias)
    test_op_mode = mc.motion.get_operation_mode(servo=alias)
    test_zero_vel = mc.communication.get_register(HOMING_ZERO_VELOCITY_REGISTER, servo=alias)
    test_sensor_index = mc.communication.get_register(
        HOMING_INDEX_PULSE_SOURCE_REGISTER, servo=alias
    )
    assert test_offset == homing_offset
    assert test_timeout == homing_timeout
    if direction == 1:
        assert test_hom_mode == HomingMode.POSITIVE_IDX_PULSE
    elif direction == 0:
        assert test_hom_mode == HomingMode.NEGATIVE_IDX_PULSE
    assert test_op_mode == OperationMode.HOMING
    assert pytest.approx(zero_vel) == test_zero_vel
    assert test_sensor_index == sensor_index
    if motor_enable:
        resolution = mc.configuration.get_position_feedback_resolution(servo=alias)
        actual_position = mc.motion.get_actual_position(servo=alias)
        assert (
            pytest.approx(homing_offset, abs=resolution * RELATIVE_ERROR_ALLOWED) == actual_position
        )


@pytest.mark.usefixtures("initial_position")
@pytest.mark.parametrize("direction", [1, 0])
def test_homing_on_switch_limit_and_index_pulse(motion_controller, direction):
    mc, alias = motion_controller
    homing_offset = 300
    homing_timeout = 3000
    search_vel = 5.0
    zero_vel = 7.0
    switch = 3
    sensor_index = 1
    mc.configuration.homing_on_switch_limit_and_index_pulse(
        homing_offset,
        direction,
        switch,
        sensor_index,
        homing_timeout,
        search_vel,
        zero_vel,
        servo=alias,
        motor_enable=False,
    )
    test_offset = mc.communication.get_register(HOMING_OFFSET_REGISTER, servo=alias)
    test_timeout = mc.communication.get_register(HOMING_TIMEOUT_REGISTER, servo=alias)
    test_hom_mode = mc.communication.get_register(HOMING_MODE_REGISTER, servo=alias)
    test_op_mode = mc.motion.get_operation_mode(servo=alias)
    test_search_vel = mc.communication.get_register(HOMING_SEARCH_VELOCITY_REGISTER, servo=alias)
    test_zero_vel = mc.communication.get_register(HOMING_ZERO_VELOCITY_REGISTER, servo=alias)
    switch_register = (
        POSITIVE_HOMING_SWITCH_REGISTER if direction == 1 else NEGATIVE_HOMING_SWITCH_REGISTER
    )
    test_switch = mc.communication.get_register(switch_register, servo=alias)
    test_sensor_index = mc.communication.get_register(
        HOMING_INDEX_PULSE_SOURCE_REGISTER, servo=alias
    )
    assert test_offset == homing_offset
    assert test_timeout == homing_timeout
    if direction == 1:
        assert test_hom_mode == HomingMode.POSITIVE_LIMIT_SWITCH_IDX_PULSE
    elif direction == 0:
        assert test_hom_mode == HomingMode.NEGATIVE_LIMIT_SWITCH_IDX_PULSE
    assert test_op_mode == OperationMode.HOMING
    assert pytest.approx(zero_vel) == test_zero_vel
    assert pytest.approx(search_vel) == test_search_vel
    assert test_switch == switch
    assert test_sensor_index == sensor_index
